from celery.utils.log import get_task_logger

from firexapp.broker_manager.broker_factory import BrokerFactory
from firexapp.engine.logging import add_hostname_to_log_records, add_custom_log_levels, PRINT_LEVEL_NAME
from firexapp.plugins import get_plugin_module_list
from firexapp.discovery import find_firex_task_bundles

add_custom_log_levels()
add_hostname_to_log_records()

# logging formats
timestamp_format = "<small>[%(asctime)s]"
process_format = "[%(levelname)s/%(processName)-13s]"
task_format = "[%(task_id).8s-%(task_name)s]"
message_format = ":</small> %(message)s"
worker_log_format = timestamp_format + process_format + message_format
worker_task_log_format = timestamp_format + process_format + task_format + message_format

logger = get_task_logger(__name__)


broker_url = BrokerFactory.get_broker_url()
result_backend = broker_url

# find default tasks
logger.debug("Beginning bundle discovery")
bundles = find_firex_task_bundles()
logger.debug("Bundle discovery completed.")
if bundles:
    logger.debug('Bundles discovered:\n' + '\n'.join([f'\t - {b}' for b in bundles]))

imports = tuple(bundles) + tuple(get_plugin_module_list()) + tuple(["firexapp.tasks.example",
                                                                    "firexapp.tasks.core_tasks",
                                                                    "firexapp.tasks.root_tasks",
                                                                    "firexapp.submit.report_trigger",
                                                                    "firexapp.reporters.json_reporter"
                                                                    ])

root_task = "firexapp.tasks.root_tasks.RootTask"

accept_content = ['pickle', 'json']
task_serializer = 'pickle'
result_serializer = 'pickle'
result_expires = None

task_track_started = True
task_acks_late = True

worker_prefetch_multiplier = 1

worker_redirect_stdouts_level = PRINT_LEVEL_NAME

mc = BrokerFactory.get_hostname_port_from_url(broker_url)[0]


import os
import sys
import inspect
from argparse import ArgumentParser, Action

from celery.signals import worker_init
from celery.utils.log import get_task_logger
from firexapp.common import delimit2list
from firexkit.task import REPLACEMENT_TASK_NAME_POSTFIX

logger = get_task_logger(__name__)
PLUGGING_ENV_NAME = "firex_external"


def get_short_name(long_name: str) -> str:
    return long_name.split('.')[-1]


def find_plugin_file(file_path):
    # is it a full path?
    if os.path.isabs(file_path):
        return file_path

    # is it relative to the cwd?
    in_cwd = os.path.abspath(file_path)
    if os.path.isfile(in_cwd):
        return in_cwd
    raise FileNotFoundError(file_path)


def cdl2list(external_files):
    if not external_files:
        return []

    external_modules = [file.strip() for file in external_files.split(",")]
    external_modules = [find_plugin_file(file) for file in external_modules if file]
    return external_modules


def get_plugin_modules(external_files):
    external_modules = cdl2list(external_files)
    if not external_modules:
        return []

    modules = []
    py_paths_to_add = []
    for file_path in external_modules:
        module_directory = os.path.dirname(os.path.abspath(file_path))
        py_paths_to_add.append(module_directory)  # allow dups. They can be removed later
        module_name = os.path.splitext(os.path.basename(file_path))[0]
        modules.append(module_name)

    # the last external takes precedence, so append those python paths first
    for p in reversed(py_paths_to_add):
        if p not in sys.path:
            sys.path.append(p)

    return modules


# noinspection PyUnusedLocal
@worker_init.connect()
def _worker_init_signal(*args, **kwargs):
    _unregister_duplicate_tasks()
    _mark_plugin_module_tasks()


def _mark_plugin_module_tasks():
    from celery import current_app
    ext_mods = get_plugin_module_list()
    for ext_mod in ext_mods:
        ext_mod_tasks = [t for t in current_app.tasks if t.startswith(ext_mod)]
        for ext_mod_task in ext_mod_tasks:
            current_app.tasks[ext_mod_task].from_plugin = True


# there is no way of copying the signals without coupling with the internals of celery signals
# noinspection PyProtectedMember
def _get_signals_with_connections():
    from celery.utils.dispatch.signal import Signal, NONE_ID
    import celery.signals as sigs
    # get all official signals
    signals = [s for s in sigs.__dict__.values() if type(s) is Signal]
    # only use the ones registered to specific microservices (as opposed to sender=None)
    signals = [s for s in signals if len(s.receivers) > len(s._live_receivers(None))]

    # now get the task specific registrations
    def from_sender_only(sig):
        return [k for k in sig.receivers if k[0][1] != NONE_ID]
    signals = {s: from_sender_only(s) for s in signals}
    signals = {s: k for s, k in signals.items() if k}

    return signals


def create_replacement_task(original, name_postfix, sigs):
    new_name = original.name + name_postfix
    bound = inspect.ismethod(original.undecorated)
    func = original.run if not bound else original.run.__func__
    options = {key: getattr(original, key) for key in ["acks_late",
                                                       "default_retry_delay",
                                                       "expires",
                                                       "ignore_result",
                                                       "max_retries",
                                                       "reject_on_worker_lost",
                                                       "resultrepr_maxsize",
                                                       "soft_time_limit",
                                                       "store_errors_even_if_ignored",
                                                       "soft_time_limit",
                                                       "time_limit",
                                                       "track_started",
                                                       "trail",
                                                       "typing",
                                                       "returns",
                                                       "flame",
                                                       "pending_child_strategy"] if key in dir(original)}

    from celery import current_app
    new_task = current_app.task(name=new_name,
                                bind=bound,
                                base=inspect.getmro(original.__class__)[1],
                                check_name_for_override_posfix=False,
                                **options)(fun=func)
    if hasattr(original, "orig"):
        new_task.orig = original.orig
    if hasattr(original, "report_meta"):
        new_task.report_meta = original.report_meta

    try:
        # there is no way of copying the signals without coupling with the internals of celery signals
        # noinspection PyProtectedMember
        from celery.utils.dispatch.signal import _make_id
        orig_task_id = _make_id(original)
        for s, receivers in sigs.items():
            for r in receivers:
                # format is ((id(receiver), id(sender)), ref(receiver))
                # locate any registered signal against the original microservice
                if r[0][1] == orig_task_id:
                    # new entry only replaces the
                    entry = ((r[0][0], _make_id(new_task)), r[1])
                    s.receivers.append(entry)
    except Exception as e:
        logger.error("Unable to copy signals while overriding %s:\n%s" % (original.name, str(e)))
    return new_task


def _unregister_duplicate_tasks():
    sigs = _get_signals_with_connections()
    from celery import current_app
    becomes = identify_duplicate_tasks(current_app.tasks, get_plugin_module_list())
    for substitutions in becomes:
        prime_overrider = substitutions[-1]  # the last item in the list is the last override
        for index in range(0, len(substitutions)-1):
            original_name = substitutions[index]
            original = current_app.tasks[original_name]
            current_app.tasks[original_name] = current_app.tasks[prime_overrider]

            postfix = REPLACEMENT_TASK_NAME_POSTFIX * (len(substitutions) - index - 1)
            new_task = create_replacement_task(original, postfix, sigs)
            overrider = substitutions[index+1]
            current_app.tasks[overrider].orig = new_task


def identify_duplicate_tasks(all_tasks, priority_modules: list) -> [[]]:
    """
    Returns a list of substitution. Each substitution is a list of microservices. The last will be the 'dominant' one.
    It will be the one used.
    """
    unique_names = set([get_short_name(long_name) for long_name in all_tasks])
    unique_names = {name: list() for name in unique_names}
    for long_name in all_tasks:
        unique_names[get_short_name(long_name)].append(long_name)

    def priority_index(micro_name):
        try:
            return priority_modules.index(os.path.splitext(micro_name)[0])
        except ValueError:
            return -1
    overrides = [sorted(long_names, key=priority_index) for long_names in unique_names.values() if len(long_names) > 1]
    return overrides


def get_plugin_module_list(external_files=None):
    if external_files is None:
        external_files = get_active_plugins()

    external_modules = get_plugin_modules(external_files)
    return external_modules


def load_plugin_modules(external_files=None):
    if external_files is None:
        external_files = get_active_plugins()
    else:
        set_plugins_env(external_files)

    external_modules = get_plugin_modules(external_files)
    if not external_modules:
        return

    external_files = cdl2list(external_files)
    for module_name in external_modules:
        import_plugin_module(module_name=module_name, external_files=external_files)
    _unregister_duplicate_tasks()


def import_plugin_module(module_name, external_files):
    __import__(module_name)
    if module_name in sys.modules:
        module_source = sys.modules[module_name].__file__
        if module_source in external_files:
            print("External module %s imported" % module_name)
        else:
            logger.error("External module %s was NOT imported. "
                         "A module with the same name was already imported from %s" % (module_name, module_source))
    else:
        logger.error("External module %s was NOT imported." % module_name)


def set_plugins_env(external_files):
    if external_files:
        external_files = cdl2list(external_files)
        os.environ[PLUGGING_ENV_NAME] = ",".join(external_files)


def get_active_plugins():
    return os.environ.get(PLUGGING_ENV_NAME, "")


def merge_plugins(*plugin_lists) -> []:
    """Merge  comma delimited lists of plugins into a single list. Right-handed most significant plugin"""
    combined_list = []
    for plugin_list in plugin_lists:
        combined_list += delimit2list(plugin_list)
    new_list = []
    for next_idx, plugin in enumerate(combined_list, start=1):
        if plugin not in combined_list[next_idx:]:
            new_list.append(plugin)
    return new_list


class CommaDelimitedListAction(Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        self.is_default = True
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super(CommaDelimitedListAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        old_value = getattr(namespace, self.dest) if hasattr(namespace, self.dest) and not self.is_default else ""
        self.is_default = False
        if old_value:
            old_value += ","
        new_value = ",".join(merge_plugins(old_value, values))
        setattr(namespace, self.dest, new_value)


plugin_support_parser = ArgumentParser(add_help=False)
plugin_support_parser.add_argument("--external", "--plugins", '-external', '-plugins', "--plugin",
                                   help="Comma delimited list of plugins files to load",
                                   default="",
                                   dest='plugins',
                                   action=CommaDelimitedListAction)



import os

from importlib import import_module
from celery import bootsteps
from celery.signals import task_postrun
from celery.states import REVOKED, RETRY
from celery.utils.log import get_task_logger
from firexkit.chain import InjectArgs, FireXTask
from firexkit.result import get_results_upto_parent, find_unsuccessful_in_chain
from firexkit.task import flame

from firexapp.application import get_app_tasks
from firexapp.engine.celery import app

logger = get_task_logger(__name__)


# noinspection PyPep8Naming
@app.task(bind=True, returns=('chain_results', 'unsuccessful_services'))
def RootTask(self, chain, **chain_args):
    c = InjectArgs(chain=chain, **chain_args)
    for task in get_app_tasks(chain):
        c |= task.s()
    promise = self.enqueue_child(c, block=True, raise_exception_on_failure=False)
    chain_results = get_results_upto_parent(promise)
    unsuccessful_services = find_unsuccessful_in_chain(promise)
    return chain_results, unsuccessful_services


def get_configured_root_task():
    # determine the configured root_task
    root_task_long_name = app.conf.get("root_task")
    root_module_name, root_task_name = os.path.splitext(root_task_long_name)
    if root_module_name != __name__:
        # Root task has been overridden
        root_task_name = root_task_name.lstrip(".")
        root_module = import_module(root_module_name)
        return getattr(root_module, root_task_name)
    else:
        # default FireXApp root task
        return RootTask


# noinspection PyUnusedLocal
@task_postrun.connect(sender=get_configured_root_task())
def handle_firex_root_completion(sender, task, task_id, args, kwargs, **do_not_care):
    logger.info("Root task completed.")

    submit_app = kwargs.get("submit_app")
    if not submit_app:
        logger.warning("Cannot find submit_app. Root task post-run handling aborted.")
        return

    result = task.AsyncResult(task_id)
    sync = kwargs.get("sync", False)

    result_state = result.state
    if sync and result_state != REVOKED and result_state != RETRY:  # Revoked can be in retry state with celery 5.1.0
        logger.debug("Sync run has not been revoked. Cleanup skipped.")
        # Only if --sync run was revoked do we want to do the shutdown here; else it's done in firex.py
        return

    # Let this signal cause self-destruct
    submit_app.self_destruct(chain_details=(result, kwargs),
                             reason=f'Root task completion ({result.state}) detected via postrun signal.')

    logger.info("Root task post run signal completed")


class BrokerShutdown(bootsteps.StartStopStep):
    """ This celery shutdown step will cleanup redis """
    label = "Broker"

    # noinspection PyMethodMayBeStatic
    def shutdown(self, parent):
        from firexapp.submit.submit import SubmitBaseApp
        if parent.hostname.startswith(SubmitBaseApp.PRIMARY_WORKER_NAME + "@"):
            # shut down the broker
            from firexapp.broker_manager.broker_factory import BrokerFactory
            BrokerFactory.broker_manager_from_env().shutdown()
            logger.debug("Broker shut down from boot step.")
        else:
            logger.debug("Not the primary celery instance. Broker will not be shut down.")


app.steps['consumer'].add(BrokerShutdown)



from firexkit.argument_conversion import ConverterRegister
from firexkit.task import flame_collapse
from firexapp.engine.celery import app
from firexapp.tasks.root_tasks import get_configured_root_task
from firexapp.submit.reporting import ReportersRegistry
from firexkit.task import flame_collapse


__all__ = ["RunInitialReport"]


@ConverterRegister.register_for_task(get_configured_root_task())
def run_initial_reporting(kwargs):
    # only run task if necessary; generators implement pre-load overloads
    generators = ReportersRegistry.get_generators()
    generators_with_pre_run = [g for g in generators if g.__class__.__dict__.get("pre_run_report")]
    if not generators_with_pre_run:
        return

    RunInitialReport.s(**kwargs).enqueue(block=False)


# noinspection PyPep8Naming
@app.task
@flame_collapse('self')
def RunInitialReport(**kwargs):
    ReportersRegistry.pre_run_report(kwargs)



import inspect
import json
import os
import sys
from _socket import gethostname

from firexapp.fileregistry import FileRegistry
from firexapp.submit.uid import Uid

PYDEV_REGISTRY_KEY = 'pydev_debug'
FileRegistry().register_file(PYDEV_REGISTRY_KEY,
                             os.path.join(Uid.debug_dirname, 'pydev_debug.json'))


def get_pydev_debug_setup():
    frame = inspect.currentframe()
    all_frames = inspect.getouterframes(frame)
    for i in range(1, len(all_frames)):
        setup_frame = all_frames[-i]
        setup = setup_frame.frame.f_locals.get('setup')
        if setup:
            return setup, setup_frame.filename


def is_debugging():
    frame = inspect.currentframe()
    top_frame = inspect.getouterframes(frame)[-1]
    return "pydevd.py" in top_frame.filename


def is_firex_submit():
    frame = inspect.currentframe()
    while frame.f_back:
        if frame.f_code.co_name == "run_submit" and str(frame.f_code.co_filename).endswith("submit.py"):
            return frame
        frame = frame.f_back
    return None


def get_pydev_command(cmd, setup=None, pydev=None, debug_host=None):
    if not setup:
        setup, pydev = get_pydev_debug_setup()
    pydev_cmd = [sys.executable, pydev]
    client = str(setup['client'])
    client = client if "127.0.0.1" not in client else (debug_host or gethostname())
    pydev_cmd += ['--multiproc', '--client', client, '--port', str(setup['port']), '--file']
    if type(cmd) is str:
        cmd = [cmd]
    pydev_cmd += cmd
    return pydev_cmd


def is_celery():
    frame = inspect.currentframe()
    while frame.f_back:
        if str(frame.f_code.co_filename).endswith("celery/bin/celery.py"):
            return frame
        frame = frame.f_back
    return None


def store_debug_info():
    setup, pydev = get_pydev_debug_setup()
    data = {
        "pydev": pydev,
        "setup": setup,
        "debug_host": gethostname()
    }

    from firexapp.engine.celery import app
    logs_dir = app.backend.get('logs_dir').decode()
    json_path = FileRegistry().get_file(PYDEV_REGISTRY_KEY, logs_dir)

    print("Storing debugging information", file=sys.stdout)
    with open(json_path, 'w') as outfile:
        json.dump(data, outfile)


def restart_celery_in_debug():
    from firexapp.engine.celery import app
    firex_logs_dir = app.backend.get('logs_dir').decode()
    json_path = FileRegistry().get_file(PYDEV_REGISTRY_KEY, firex_logs_dir)

    if not os.path.isfile(json_path):
        print("No debugging information found", file=sys.stdout)
        return

    with open(json_path, "r") as infile:
        data = json.load(infile)

    setup = data['setup']
    pydev = data['pydev']
    debug_host = data['debug_host']

    celery_argsv = sys.argv
    debug_celery = get_pydev_command(celery_argsv, setup=setup, pydev=pydev, debug_host=debug_host)
    print("restarting celery in debug mode", file=sys.stdout)
    os.execl(debug_celery[0], *debug_celery)


if is_debugging():
    firex_frame = is_firex_submit()
    if firex_frame:
        # store the debug info to be used when celery restarts
        store_debug_info()
elif is_celery():
    restart_celery_in_debug()



import os
import abc
import sys
import inspect
from importlib import import_module


class FlowTestConfiguration(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.results_folder = ""

    @property
    def name(self):
        return self.__class__.__name__

    @abc.abstractmethod
    def initial_firex_options(self)->list:
        pass

    @abc.abstractmethod
    def assert_expected_firex_output(self, cmd_output, cmd_err):
        pass

    @abc.abstractmethod
    def assert_expected_return_code(self, ret_value):
        pass


class InterceptFlowTestConfiguration(FlowTestConfiguration):
    __metaclass__ = abc.ABCMeta

    def assert_expected_return_code(self, ret_value):
        assert 0 == ret_value, "Expected return code of 0"

    def assert_expected_firex_output(self, cmd_output, cmd_err):
        return cmd_err is None, "Intercept tests should not have errors"

    @abc.abstractmethod
    def intercept_service(self)->str:
        """
            Name of the microservice that will be mocked into the validator capturing the options that get compared by
            assertExpectedOptions()
        """
        pass

    @abc.abstractmethod
    def assert_expected_options(self, captured_kwargs):
        pass


def assert_is_bad_run(ret_value):
    assert ret_value != 0, "This test should have a FAILURE return code, but returned 0"


def assert_is_good_run(ret_value):
    assert ret_value == 0, "Test expects a CLEAN run, but returned %s. " \
                           "Check the err output to see what went wrong." % str(ret_value)


def skip_test(cls):
    setattr(cls, "skip_test", True)
    return cls


def discover_tests(tests, config_filter="") -> list:
    configs = []
    for tests_path in tests.split(","):

        if not os.path.exists(tests_path):
            print("Error: --tests must be a directory or a python module containing test configs\n"
                  "%s is not recognized" % tests_path, file=sys.stderr)
            exit(-1)
        configs += import_test_configs(tests_path)
    if config_filter:
        filters = [config_filter.strip() for config_filter in config_filter.split(",")]
        configs = [config for config in configs if config.__class__.__name__ in filters]

    [print("Skipping " + config.__class__.__name__, file=sys.stderr) for config in configs if hasattr(config,
                                                                                                      "skip_test")]
    configs = [config for config in configs if not hasattr(config, "skip_test")]

    return configs


def import_test_configs(path) -> []:
    # dynamically load module
    if not os.path.exists(path):
        raise FileNotFoundError(path)
    if (__file__ in path or
            "pycache" in path or  # We don't need to look at the cache
            os.path.basename(path) == "data"):  # By convention, a "data" directory will contain artifacts for the tests
        return []

    config_objects = []
    if os.path.isfile(path):
        if os.path.splitext(path)[1] != ".py":
            return []
        if os.path.basename(__file__) == os.path.basename(path):
            return []

        sys.path.append(os.path.dirname(os.path.abspath(path)))
        module = import_module(os.path.splitext(os.path.basename(path))[0])

        for _, obj in inspect.getmembers(module, inspect.isclass):
            if FlowTestConfiguration.__name__ in [cls.__name__ for cls in inspect.getmro(obj)[1:]] and \
                            not inspect.isabstract(obj) and '__metaclass__' not in obj.__dict__ and \
                            obj.__module__ == module.__name__:
                config_object = obj()
                config_object.filepath = path
                config_objects.append(config_object)

    elif os.path.isdir(path):
        results_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), "results")
        if os.path.normpath(path) == os.path.normpath(results_folder):
            return []
        for sub_path in [os.path.join(path, f) for f in os.listdir(path)]:
            config_objects += import_test_configs(sub_path)
    return config_objects



import os
import datetime
import pytz
import tempfile
from getpass import getuser
import random
import firexkit
import shutil

from firexapp.submit.arguments import whitelist_arguments
from firexkit.permissions import DEFAULT_CHMOD_MODE

BASE_LOGGING_DIR_ENV_VAR_KEY = 'firex_base_logging_dir'


class Uid(object):
    debug_dirname = 'firex_internal'
    _resources_dirname = os.path.join(debug_dirname, 'resources')

    def __init__(self, identifier=None):
        self.timestamp = datetime.datetime.now(tz=pytz.utc)
        self.user = getuser()
        if identifier:
            self.identifier = identifier
        else:
            random.seed()
            self.identifier = f'FireX-{self.user}-{self.timestamp.strftime("%y%m%d-%H%M%S")}-{random.randint(1, 65536)}'
        self._base_logging_dir = None
        self._logs_dir = None
        self._debug_dir = None
        self._viewers = {}

    @property
    def base_logging_dir(self):
        if not self._base_logging_dir:
            self._base_logging_dir = os.environ.get(BASE_LOGGING_DIR_ENV_VAR_KEY, tempfile.gettempdir())
        return self._base_logging_dir

    @property
    def logs_dir(self):
        if not self._logs_dir:
            self._logs_dir = self.create_logs_dir()
            self._debug_dir = self.create_debug_dir()
            self.copy_resources()
        return self._logs_dir

    @property
    def debug_dir(self):
        if not self._debug_dir:
            self._debug_dir = self.create_debug_dir()
        return self._debug_dir

    @classmethod
    def get_resources_path(cls, logs_dir):
        return os.path.join(logs_dir, cls._resources_dirname)

    @property
    def resources_dir(self):
        return self.get_resources_path(self.logs_dir)

    def create_logs_dir(self):
        path = os.path.join(self.base_logging_dir, self.identifier)
        os.makedirs(path)
        return path

    def create_debug_dir(self):
        path = os.path.join(self.logs_dir, self.debug_dirname)
        try:
            os.makedirs(path)
        except FileExistsError:
            # Could have been created by other dependencies (e.g. redis)
            pass
        return path

    def __str__(self):
        return self.identifier

    def __repr__(self):
        return self.identifier

    def __eq__(self, other):
        return str(other) == self.identifier

    def copy_resources(self):
        # pkg_resources.resource_filename('firexkit', 'resources') would have been a cleaner way, but
        # pkg_reources is very slow to load
        pkg_resource_dir = os.path.join(os.path.dirname(firexkit.__file__), 'resources')
        resources_dir = self.resources_dir
        shutil.copytree(pkg_resource_dir, resources_dir)
        # Open permissions
        os.chmod(self.resources_dir, DEFAULT_CHMOD_MODE)
        for file in os.listdir(resources_dir):
            os.chmod(os.path.join(resources_dir, file), DEFAULT_CHMOD_MODE)

    def add_viewers(self, **attrs):
        self._viewers.update(**attrs)

    @property
    def viewers(self):
        return self._viewers

    @property
    def logs_url(self):
        try:
            return self.viewers['logs_url']
        except KeyError:
            return None


whitelist_arguments("uid")



import os
import sys
import logging
from collections import OrderedDict, namedtuple
from typing import Dict, List, Tuple

from entrypoints import EntryPoint

TASKS_DIRECTORY = "firex_tasks_directory"

logger = logging.getLogger(__name__)

_loaded_firex_bundles = {}


class PkgVersionInfo(namedtuple('PkgVersionInfo', ('pkg', 'version', 'commit'), defaults=(None, None, None))):
    def __str__(self):
        return f'{self.pkg}: {self.version or self.commit}'


def _get_paths_without_cwd() -> [str]:
    # This is needed because Celery temporarily adds the cwd into the sys.path via a context switcher,
    # and our discovery takes place inside that context.
    # Having cwd in the sys.path can slow down the discovery significantly without any benefit.
    paths = list(sys.path)
    try:
        paths.remove(os.getcwd())
    except ValueError:  # pragma: no cover
        pass
    return paths


#
# In case there are duplicate modules found, only keep one for each
#   (name, module_name, object_name) tuple. This prevents duplicate
#   arg registration failures when the sys.path causes the same service
#   to be found twice.
#
def prune_duplicate_module_entry_points(entry_points) -> [EntryPoint]:
    id_to_entry_points = OrderedDict()

    for e in entry_points:
        key = (e.name, e.module_name, e.object_name)
        if key not in id_to_entry_points:
            id_to_entry_points[key] = e
        # Replace the currently stored entry point for this key if the distro is None.
        elif id_to_entry_points[key].distro is None and e.distro is not None:
            id_to_entry_points[key] = e

    return list(id_to_entry_points.values())


def _get_entrypoints(name, prune_duplicates=True, path=None) -> [EntryPoint]:
    import entrypoints
    if path is not None and not isinstance(path, list):
        path = [path]
    eps = [ep for ep in entrypoints.get_group_all(name, path=path)]
    if prune_duplicates:
        eps = prune_duplicate_module_entry_points(eps)
    return eps


def get_firex_bundles_entry_points(path=None) -> [EntryPoint]:
    return _get_entrypoints('firex.bundles', path=path)


def loaded_firex_bundles_entry_points(path=None) -> Dict[EntryPoint, object]:
    global _loaded_firex_bundles
    key = str(path)
    if key not in _loaded_firex_bundles:
        eps = get_firex_bundles_entry_points(path=path)
        _loaded_firex_bundles[key] = {ep: ep.load() for ep in eps}
    return _loaded_firex_bundles[key]


def get_firex_tracking_services_entry_points() -> [EntryPoint]:
    return _get_entrypoints('firex_tracking_service')


def get_firex_dependant_package_versions() -> [PkgVersionInfo]:
    versions = list()
    for ep, loaded_pkg in loaded_firex_bundles_entry_points().items():
        try:
            version = loaded_pkg.__version__
        except AttributeError:
            version = None
        try:
            commit = loaded_pkg._version.get_versions()['full-revisionid']
        except AttributeError:
            commit = None
        versions.append(PkgVersionInfo(pkg=ep.name, version=version, commit=commit))
    return versions


def get_all_pkg_versions() -> [PkgVersionInfo]:
    import firexkit
    import firexapp
    from firexapp.submit.tracking_service import get_tracking_services_versions
    return [
               PkgVersionInfo(pkg='firexkit',
                              version=firexkit.__version__,
                              commit=firexkit._version.get_versions()['full-revisionid']),
               PkgVersionInfo(pkg='firexapp',
                              version=firexapp.__version__,
                              commit=firexapp._version.get_versions()['full-revisionid']),
    ] + get_tracking_services_versions() + get_firex_dependant_package_versions()


def get_all_pkg_versions_as_dict() -> dict():
    return {pkg_info.pkg: pkg_info for pkg_info in get_all_pkg_versions()}

def _find_bundle_pkg_root(path, namespace):
    while True:
        head, tail = os.path.split(path)
        if tail == namespace:
            return head
        else:
            path = os.path.dirname(path)


# Return a list of two-element tuples
# Where the 1st element is the path of package
# and the 2nd element is the path of package's root
def _get_firex_bundle_package_locations(path=None) -> List[Tuple[str, str]]:
    locations = []
    loaded_entry_points = loaded_firex_bundles_entry_points(path=path)
    for p in loaded_entry_points.values():
        namespace = p.__package__.split('.')[0]
        pkg_paths = p.__path__
        for pkg_path in pkg_paths:
            root = _find_bundle_pkg_root(pkg_path, namespace)
            locations.append((pkg_path, root))
    return locations


def discover_package_modules(current_path, root_path=None) -> [str]:
    if root_path is None:
        root_path = os.path.dirname(current_path)

    services = []
    if os.path.isfile(current_path):
        basename, ext = os.path.splitext(current_path)
        if ext.lower() == ".py" and not os.path.basename(current_path).startswith('_'):
            basename = basename.replace(root_path, "")
            return [basename.replace(os.path.sep, ".").strip(".")]
        else:
            return []
    elif os.path.isdir(current_path):
        base = os.path.basename(current_path)
        if "__pycache__" in base or base.startswith("."):
            return []
        for child_name in os.listdir(current_path):
            full_child = os.path.join(current_path, child_name)
            services += discover_package_modules(full_child, root_path)
        return services
    else:
        # either a symlink or a path that doesn't exist
        return []


def find_firex_task_bundles() -> [str]:
    # look for task modules in dependant packages
    bundles = []
    locations = _get_firex_bundle_package_locations()
    for path, root_path in locations:
        bundles += discover_package_modules(path, root_path)
    # look for task modules in env defined location
    if TASKS_DIRECTORY in os.environ:
        include_location = os.environ[TASKS_DIRECTORY]
        if os.path.isdir(include_location):
            if include_location not in sys.path:
                sys.path.append(include_location)
            include_tasks = discover_package_modules(include_location, root_path=include_location)
            bundles += include_tasks

    return bundles



import time

import os
import psutil
import re
import socket

from jinja2 import Template
from celery.utils.log import get_task_logger

logger = get_task_logger(__name__)

FIREX_BIN_DIR_ENV = 'firex_bin_dir'


def delimit2list(str_to_split, delimiters=(',', ';', '|', ' ')) -> []:
    if not str_to_split:
        return []

    if not isinstance(str_to_split, str):
        return str_to_split

    # regex for only comma is (([^,'"]|"(?:\\.|[^"])*"|'(?:\\.|[^'])*')+)
    regex = """(([^""" + "".join(delimiters).replace(" ", "\s") + """'"]|"(?:\\.|[^"])*"|'(?:\\.|[^'])*')+)"""
    tokens = re.findall(regex, str_to_split)

    # unquote "tokens" if necessary
    tokens = [g1 if g1.strip() != g2.strip() or g2[0] not in "'\"" else g2.strip(g2[0]) for g1, g2 in tokens]
    tokens = [t.strip() for t in tokens]  # remove extra whitespaces
    tokens = [t.strip("".join(delimiters) + " ") for t in tokens]  # remove any extra (or lone) delimiters
    tokens = [t for t in tokens if t]  # remove empty tokens
    return tokens


def get_available_port(so_reuseport=True):
    with socket.socket() as sock:
        if so_reuseport and hasattr(socket, 'SO_REUSEPORT'):
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(('', 0))
        port = sock.getsockname()[1]
    return port


def silent_mkdir(path, exist_ok=True, **kwargs):
    os.makedirs(path, exist_ok=exist_ok, **kwargs)


def poll_until_path_exist(path, timeout=10):
    timeout_time = time.time() + timeout
    path_exists = os.path.exists(path)
    while not path_exists and time.time() < timeout_time:
        time.sleep(0.1)
        path_exists = os.path.exists(path)
    if not path_exists:
        raise AssertionError(f'{path} did not exist within {timeout}s')


def poll_until_file_exist(file_path, timeout=10):
    poll_until_path_exist(file_path, timeout=timeout)
    assert os.path.isfile(file_path), f'{file_path} does not appear to be a file'


def poll_until_existing_file_not_empty(file_path, timeout=10):
    timeout_time = time.time() + timeout
    while os.stat(file_path).st_size == 0 and time.time() < timeout_time:
        time.sleep(0.1)
    assert os.stat(file_path).st_size > 0, 'File %s size is zero' % file_path


def poll_until_file_not_empty(file_path, timeout=10):
    start_time = time.time()
    poll_until_file_exist(file_path, timeout)
    remaining_timeout = timeout - (time.time() - start_time)
    poll_until_existing_file_not_empty(file_path, remaining_timeout)


def poll_until_dir_empty(dir_path, timeout=15):
    timeout_time = time.time() + timeout
    while len(os.listdir(dir_path)) > 0 and time.time() < timeout_time:
        time.sleep(0.1)
    return not os.listdir(dir_path)


def proc_matches(proc_info, pname, cmdline_regex, cmdline_contains):
    if proc_info['name'] == pname:
        if cmdline_regex:
            return any(cmdline_regex.search(item) for item in proc_info['cmdline'])
        elif cmdline_contains:
            return any(cmdline_contains in item for item in proc_info['cmdline'])
        else:
            return True
    else:
        return False


def find_procs(name, cmdline_regex=None, cmdline_contains=None):
    matching_procs = []
    if cmdline_regex:
        cmdline_regex = re.compile(cmdline_regex)
    else:
        cmdline_regex = None
    for proc in psutil.process_iter():
        try:
            pinfo = proc.as_dict(attrs=['name', 'cmdline', 'pid'])
        except psutil.NoSuchProcess:
            pass
        else:
            if proc_matches(pinfo, name, cmdline_regex, cmdline_contains):
                matching_procs.append(proc)

    return matching_procs


def wait_until(predicate, timeout, sleep_for, *args, **kwargs):
    max_time = time.time() + timeout
    while time.time() < max_time:
        if predicate(*args, **kwargs):
            return True
        time.sleep(sleep_for)
    return predicate(*args, **kwargs)


def wait_until_pid_not_exist(pid, timeout=7, sleep_for=1):
    return wait_until(lambda p: not psutil.pid_exists(p), timeout, sleep_for, pid)


def qualify_firex_bin(bin_name):
    if FIREX_BIN_DIR_ENV in os.environ:
        return os.path.join(os.environ[FIREX_BIN_DIR_ENV], bin_name)
    return bin_name


def select_env_vars(env_names):
    return {k: v for k, v in os.environ.items() if k in env_names}


def find(keys, input_dict):
    result = input_dict
    for key in keys:
        try:
            result = result[key]
        except Exception:
            return None
    return result


def render_template(template_str, template_args):
    return Template(template_str).render(**template_args)


#
# Create a symlink from src -> target.
#
# delete_link == True: Delete target link first if it exists.
#             == False: Don't delete target link if it exists.
#             == None (default): If symlink creation fails because the link
#                                exists, delete it and try again.
#                                (optimized for cases where we don't expect
#                                the link to exist in most cases.)
#
def create_link(src, target, delete_link=None, relative=False):
    done = False
    attempts = 0

    if relative:
        src = os.path.relpath(src, os.path.dirname(target))

    while not done:
        if delete_link is True:
            try:
                os.remove(target)
            except Exception:
                pass
        attempts += 1
        try:
            os.symlink(src, target)
            done = True
            logger.debug('Symbolic link created: %s -> %s' % (src, target))
        except FileExistsError as e:
            if delete_link is False:
                raise
            if attempts > 1:
                # Maybe we're competing with someone else
                # to create that link...
                logger.debug('Symbolic link failed: %s -> %s' % (src, target))
                logger.warning(e)
                done = True
            else:
                # Remove target and try again
                delete_link = True


def dict2str(mydict, sort=False, sep='    ', usevrepr=True, line_prefix=''):
    if not mydict:
        return 'None'

    txt = ''
    items = mydict.items()
    if sort:
        items = sorted(items)
    maxlen = len(max(mydict.keys(), key=len))
    wrap_space = '\n' + ' ' * (maxlen + len(sep))
    for k, v in items:
        txt += line_prefix
        if usevrepr:
            txt += '%s%s%r\n' % (k.ljust(maxlen, " "), sep, v)
        else:
            v = str(v)
            v = v.replace('\n', wrap_space)
            txt += '%s%s%s\n' % (k.ljust(maxlen, " "), sep, v)
    return txt


import inspect
import re
from textwrap import wrap
from celery.exceptions import NotRegistered
from firexapp.plugins import plugin_support_parser
from firexapp.application import import_microservices, get_app_task


class InfoBaseApp:
    def __init__(self):
        self._list_sub_parser = None
        self._info_sub_parser = None

    def create_list_sub_parser(self, sub_parser):
        list_parser = sub_parser.add_parser("list", help="Lists FireX microservices, or used arguments"
                                                         "  {microservices,arguments}",
                                            parents=[plugin_support_parser])
        list_group = list_parser.add_mutually_exclusive_group(required=True)
        list_group.add_argument("--microservices", '-microservices', help="Lists all available microservices",
                                action='store_true')
        list_group.add_argument("--arguments", '-arguments', help="Lists all arguments used by microservices",
                                action='store_true')

        list_group.set_defaults(func=self.run_list)

        return list_group

    def create_info_sub_parser(self, sub_parser):
        if not self._info_sub_parser:

            info_parser = sub_parser.add_parser("info", help="Lists detailed information about a microservice",
                                                parents=[plugin_support_parser])
            info_parser.add_argument("entity", help="The short or long name of the microservice to be detailed, or a "
                                                    "microservice argument. It can be a Python compatible regexp to "
                                                    "display information about all services matching that expression.")

            info_parser.set_defaults(func=self.run_info)
            self._info_sub_parser = info_parser
        return self._info_sub_parser

    def run_list(self, args):
        if args.microservices:
            self.print_available_microservices(args.plugins)
        elif args.arguments:
            self.print_argument_used(args.plugins)

    def run_info(self, args):
        self.print_details(args.entity, args.plugins)

    @staticmethod
    def print_available_microservices(plugins: str):
        apps = import_microservices(plugins)
        print()
        print("The following microservices are available:")

        services = [str(task) for task in apps]
        services = [task for task in services if not task.startswith('celery.')]  # filter out base celery types
        services.sort()
        for service in services:
            print(service)

        pointers = [(full, apps[full].name) for full in apps if apps[full].name not in full]
        if pointers:
            print("\nPointers (override -> original):")
            for new, old in pointers:
                print(new, "->", old)
        print("\nUse the info sub-command for more details\n")

    @staticmethod
    def print_argument_used(plugins: str):
        all_tasks = import_microservices(plugins)
        print()
        print("The following arguments are used by microservices:")
        usage = get_argument_use(all_tasks)
        for arg in sorted(usage):
            print(arg)
        print("\nUse the info sub-command for more details\n")

    def print_partial_task_matches(self, entity, all_tasks):
        entries_found = False
        for task_name in sorted(all_tasks, key=lambda i: i.split('.')[-1]):
            # Is this even a partial match
            if not re.search(entity, task_name):
                continue
            try:
                task = get_app_task(task_name, all_tasks)
            except NotRegistered:
                continue
            else:
                if entries_found:
                    print('\n')
                self.print_task_details(task)
                entries_found = True
        return entries_found

    def print_details(self, entity, plugins, all_tasks=None):
        if not all_tasks:
            all_tasks = import_microservices(plugins)

        # Is this entity a microservice
        try:
            # Do we have a match on the full name?
            task = get_app_task(entity, all_tasks)
        except NotRegistered:
            # Do we have a match on the partial name
            if self.print_partial_task_matches(entity, all_tasks):
                return
        else:
            self.print_task_details(task)
            return

        # Is this entity an argument
        all_args = get_argument_use(all_tasks)
        if entity in all_args:
            print("Argument name: " + entity)
            print("Used in the following microservices:")
            for micro in all_args[entity]:
                print(micro.name)
            return

        self._info_sub_parser.exit(status=-1, message="Microservice %s was not found!" % entity)

    @classmethod
    def parse_task_docstring(cls, task):
        if not task.__doc__:
            return None, None

        header = None
        arg_dict = None
        docstring = inspect.getdoc(task)

        match = re.search(r"^(.*)\n\s*Arguments?[^\n]*\n\s*-*(.*)", docstring, re.MULTILINE | re.DOTALL)
        if match:
            if len(match.group(1).strip()):
                header = match.group(1).strip()
            if len(match.group(2).strip()):
                arg_desc_str = match.group(2).strip()
                arg_dict = {}

                # Need to determine if args are indicated with prefixed '--'
                if re.search("^\s*--", arg_desc_str):
                    # assume our args all start with --
                    arg_prefix = "(?:--)"
                else:
                    # no prefix
                    arg_prefix = ""
                # Go over arguments section and create dict of args/description
                desc = ''
                arg = None
                for line in arg_desc_str.split("\n"):
                    line = line.strip()
                    # blank line ends entry
                    if not len(line):
                        if arg:
                            arg_dict[arg] = desc
                        arg = None
                        desc = ''
                    # Look for start of new entry
                    match = re.search(r"^" + arg_prefix + r"(\S[^:\(\r\n]+)(?:\([^)\r\n]*\))?:(?:(?:\([^)\r\n]*\))?[^\S\r\n]*([^\r\n]*))?$", line)
                    if match:
                        if arg:
                            arg_dict[arg] = desc
                        arg = match.group(1)
                        desc = match.group(2).strip()
                        # Remove ':' and any leading () that usually indicate type. These can appear in either order
                        match = re.search(r"^(?:\([^)]*\)\s*)?:(?:\([^)]*\))?\s*(.+)$", desc)
                        if match:
                            desc = match.group(1)
                    else:
                        # continue prev entry
                        if len(desc):
                            desc += " " + line
                        else:
                            desc = line
                # store last outstanding entry
                if arg:
                    arg_dict[arg] = desc
        else:
            header = docstring
        return header, arg_dict

    @classmethod
    def print_task_details(cls, task):
        dash_length = 40
        print('-' * dash_length)
        split_name = task.name.split(".")
        name = split_name[-1]
        path = '.'.join(split_name[0:-1])
        if path:
            path = " (%s)" % path
        print("Name: " + name + path)

        header, arguments = cls.parse_task_docstring(task)

        if header:
            print('\n' + header)
        if not arguments:
            arguments = {}

        def print_arg(arg, deflt, description):
            max_arg_len = 25
            max_desc_len = 80 - max_arg_len - len(tab)
            arg_str = f"{tab}{arg}"

            # Add default value, if present
            if deflt is not None:
                arg_str += f"(default={deflt})"

            if description:
                # Add filler or newline, depending on arg_str length
                if len(arg_str) >= max_arg_len:
                    arg_str += "\n" + (' ' * max_arg_len)
                else:
                    arg_str += " " * (max_arg_len - len(arg_str))

                # Remove excess spacing in description
                description = re.sub(r" {2,}", " ", description)

                # Add description
                if len(description) < max_desc_len:
                    arg_str += description
                else:
                    arg_str += f"\n{' ' * max_arg_len}".join(wrap(description, max_desc_len))
            print(arg_str)

        tab = ' ' * 1
        print("\nArguments info")
        print("--------------")
        required_args = getattr(task, "required_args", [])
        cnt = 0
        for chain_arg in sorted(required_args):
            if "self" not in chain_arg and \
               "uid" not in chain_arg and \
               chain_arg != 'kwargs':
                desc = arguments.get(chain_arg, None)
                print_arg(chain_arg, None, desc)
                cnt += 1

        optional_args = getattr(task, "optional_args", {})
        if len(optional_args):
            for chain_arg in sorted(optional_args):
                if chain_arg.startswith('_'):
                    continue
                desc = arguments.get(chain_arg, None)
                default = optional_args[chain_arg]
                if default is None:
                    default = "None"
                print_arg(chain_arg, default, desc)
        else:
            if not cnt:
                print(tab, "None")

        print('\nReturns\n-------')
        out = getattr(task, "return_keys", {})
        if out:
            for chain_arg in sorted(out):
                desc = arguments.get(chain_arg, None)
                print_arg(chain_arg, None, desc)
        else:
            print(tab, "None")


def get_argument_use(all_tasks) -> dict:
    argument_usage = {}
    for _, task in all_tasks.items():
        if not hasattr(task, "required_args") or not hasattr(task, "optional_args"):
            continue

        for arg in task.required_args + list(task.optional_args):
            if arg in argument_usage:
                argument_usage[arg].add(task)
            else:
                tasks = set()
                tasks.add(task)
                argument_usage[arg] = tasks

    return argument_usage



import sys

import os
import shutil
import argparse
import unittest
from firexapp.testing.config_base import discover_tests
from firexapp.testing.config_interpreter import ConfigInterpreter
from firexkit.permissions import DEFAULT_UMASK

TEST_EXE = os.path.realpath(os.path.abspath(__file__))
if not os.path.isfile(TEST_EXE):
    raise Exception("Some import changed the cwd. Can't locate relative files")
TEST_EXE_DIR = os.path.dirname(TEST_EXE)


class FlowTestInfra(unittest.TestCase):
    test_configs = []
    results_dir = None
    failures = 0
    max_acceptable_failures = None
    config_interpreter = ConfigInterpreter()

    @classmethod
    def populate_tests(cls):
        if not cls.results_dir:
            raise Exception("Results directory not set")
        if not cls.test_configs:
            raise Exception("Could not generate tests. No configurations defined\n"
                            "Configs: %s" % str(cls.test_configs))

        # noinspection PyTypeChecker
        cls.max_acceptable_failures = int((len(cls.test_configs) / 2.0) + 1)
        print("Total tests: " + str(len(cls.test_configs)), file=sys.stderr)
        for test_config in cls.test_configs:
            def sub_test(self, config=test_config):
                    print("Running tests " + config.__class__.__name__, file=sys.stderr)
                    try:
                        cls.config_interpreter.run_integration_test(config, self.results_dir)
                        self.assertTrue(True)
                        print("\tPassed", file=sys.stderr)
                    except Exception as e:
                        print("\tFailed", file=sys.stderr)
                        cls.failures += 1
                        raise e

            setattr(cls, 'test_' + test_config.__class__.__name__, sub_test)

    def setUp(self):
        self._outcome.result.dots = False

    def tearDown(self):
        if self.failures > self.max_acceptable_failures:
            print("-"*70, file=sys.stderr)
            print("Run was terrible. Half have failed so far. Skipping the remaining test", file=sys.stderr)
            print("", file=sys.stderr)
            self._outcome.result.shouldStop = True


def main(default_results_dir, default_test_dir):
    parser = argparse.ArgumentParser()
    parser.add_argument("--logs", '-l', help="The directory to store results and mocks",
                        default=default_results_dir)
    parser.add_argument("--tests", "--test", '-t', dest="tests",
                        help="The directory or python module containing the test configurations. "
                             "Supports comma delimited lists",
                        default=default_test_dir)
    parser.add_argument("--config", "--configs", '-c', dest="config",
                        help="A comma separated list of test configurations to run", default=None)
    parser.add_argument("--xunit_file_name", help="Name of the xunit file", default=None)
    extras = parser.add_mutually_exclusive_group()
    extras.add_argument("--profile", action='store_true', help="Turn on profiling")
    extras.add_argument("--coverage", action='store_true', help="Turn on code coverage. A .coverage file will be "
                                                                "generated in the logs directory")
    parser.add_argument("--no_html", action='store_true', help="Do not generate an html code coverage report. "
                                                               "Used in combination with --coverage")
    parser.add_argument("--public_runs", action='store_true', default=False,
                        help="Should links be generated to point to public flame deployment?")
    args = parser.parse_args()

    if args.coverage:
        # coverage requires eventlet, but firexapp does not
        try:
            import eventlet
        except ModuleNotFoundError:
            print("eventlet is not installed. eventlet is necessary to get code coverage."
                  "Please run again without the --coverage option", file=sys.stderr)
            exit(-1)
    elif args.no_html:
        parser.error("--no_html cannot be used without --coverage")

    # prepare logging directory
    results_directory = os.path.realpath(args.logs)
    if os.path.isdir(results_directory):
        shutil.rmtree(results_directory)
    os.umask(DEFAULT_UMASK)
    os.mkdir(results_directory)

    FlowTestInfra.config_interpreter.profile = args.profile
    FlowTestInfra.config_interpreter.coverage = args.coverage
    FlowTestInfra.config_interpreter.is_public = args.public_runs
    FlowTestInfra.results_dir = results_directory
    FlowTestInfra.test_configs = discover_tests(args.tests, args.config)
    FlowTestInfra.populate_tests()

    # if running a single suite, rename the test to reflect the suite
    xunit_file_name = "TEST-%s.%s-results.xml" % (FlowTestInfra.__module__, FlowTestInfra.__name__)
    if os.path.isfile(args.tests):
        FlowTestInfra.__name__ = os.path.splitext(os.path.basename(args.tests))[0]
        orig_output = os.path.join(args.logs, xunit_file_name.replace("FlowTestInfra",
                                                                      FlowTestInfra.__name__))
    else:
        orig_output = os.path.join(args.logs, xunit_file_name)

    if args.coverage:
        os.environ["COVERAGE_FILE"] = os.path.join(results_directory, ".coverage")

    import xmlrunner
    success = unittest.main(module=FlowTestInfra.__module__,
                            testRunner=xmlrunner.XMLTestRunner(output=args.logs, outsuffix="results"),
                            argv=sys.argv[:1],
                            exit=False).result.wasSuccessful()

    # let the user decide what to call the output
    if args.xunit_file_name:
        os.rename(orig_output,
                  os.path.join(args.logs, os.path.basename(args.xunit_file_name)))
    if args.coverage:
        print("combining coverage files...", file=sys.stderr)
        coverage_files = [f for f in os.listdir(results_directory) if f.startswith(".coverage")]
        import subprocess
        if coverage_files:
            subprocess.check_output(["coverage", "combine"] + coverage_files,
                                    cwd=results_directory)

            if not args.no_html:
                cov_report = os.path.join(results_directory, "coverage")
                print("Generating Coverage Report...", file=sys.stderr)
                subprocess.check_output(["coverage", "html", "-d", cov_report],
                                        cwd=results_directory)
                print(cov_report, file=sys.stderr)

    sys.exit(not success)


def default_main():
    # determine default location to look for tests
    import firexapp
    module_dir = os.path.dirname(firexapp.__file__)
    root_dir = os.path.dirname(module_dir)
    package_tests_dir = os.path.join(root_dir, "tests", "integration_tests")
    if os.path.isdir(package_tests_dir):
        default_test_location = package_tests_dir
    else:
        default_test_location = "."
    main(default_results_dir=os.path.join(os.getcwd(), "results"),
         default_test_dir=default_test_location)


if __name__ == "__main__":
    default_main()



from getpass import getuser
import time

from firexkit.argument_conversion import SingleArgDecorator
from firexkit.chain import InjectArgs, returns
from firexkit.task import FireXTask, flame, flame_collapse

from firexapp.engine.celery import app
from firexapp.submit.arguments import InputConverter
from firexapp.tasks.core_tasks import CopyBogKeys


@app.task
def nop():
    return


@app.task
def sleep(sleep=None):
    if sleep:
        time.sleep(int(sleep))
    return


@app.task(returns='username')
def getusername():
    return getuser()


# The @app.task() makes this normal python function a FireX Service.
@app.task(returns=['greeting'], flame=['greeting'])
def greet(name=getuser()):
    assert len(name) > 1, "Cannot greet a name with 1 or fewer characters."
    return 'Hello %s!' % name


# Setting bind=True makes the first argument received by the service 'self'. It's most commonly used to invoke
# (enqueue) other services, but provides much more functionality as outlined here:
@app.task(bind=True, returns=['guests_greeting'], flame=['guests_greeting'])
def greet_guests(self: FireXTask, guests):
    child_promises = []
    for guest in guests:
        # Create a Celery Signature, see: https://docs.celeryproject.org/en/latest/userguide/canvas.html#signatures
        greet_signature = greet.s(name=guest)
        # enqueue_child by default schedules the supplied signature for execution asynchronously and immediately
        # returns the newly created child result promise.
        child_promise = self.enqueue_child(greet_signature)
        child_promises.append(child_promise)

    # We want to get all the return values (greetings) from child tasks, but we must wait first to make sure they're all
    # available before inspecting them with child_promise.result[<returns_key>]
    self.wait_for_children(raise_exception_on_failure=False)
    # Since wait_for_children has completed, we know it's safe to inspect the results of all child task promises,
    # after verifying the task was a success.
    greetings = [promise.result['greeting'] for promise in child_promises if promise.successful()]

    if any(promise.failed() for promise in child_promises):
        greetings.append("And apologies to those not mentioned.")

    return ' '.join(greetings)


@InputConverter.register
@SingleArgDecorator('guests')
def to_list(guests):
    return guests.split(',')


@app.task(returns=['amplified_message'])
def amplify(to_amplify, upper=True, surround_str=None, underline_char=None, overline_char=None):
    result = to_amplify
    if upper:
        result = to_amplify.upper()
    if surround_str:
        result = surround_str + result + surround_str
    centerline_len = len(result)
    if underline_char:
        result = result + '\n' + (underline_char * centerline_len)

    if overline_char:
        result = (overline_char * centerline_len) + '\n' + result

    return result


def _amplified_greeting_formatter(args_and_maybe_results):
    # Since 'amplified_greeting' is the return value name, it isn't available to the formatter when the task is first
    # started. It will be available if the task completes successfully.
    if 'amplified_greeting' in args_and_maybe_results:
        br_as_nl = args_and_maybe_results["amplified_greeting"].replace('\n', '<br>')
        return f'<h1 style="font-family: monospace;">{br_as_nl}</h1>'

    # Since 'guests' is an input argument, it will always be available to the formatter, even before the service
    # has completed.
    return f'Planning to greet: {",".join(args_and_maybe_results["guests"])}'


@app.task(bind=True, returns=['amplified_greeting'])
@flame('*', _amplified_greeting_formatter)
@flame_collapse({'greet_guests': 'descendants'})
def amplified_greet_guests(self: FireXTask, guests):
    # Nonsense failure case to illustrate flame HTML data when the service fails (i.e. no return value present).
    assert len(guests) > 1, "Only willing to amplify greeting for more than one guest."

    # Create a chain that can be enqueued. The greet_guests service will produce a guests_greeting,
    # which will then be delivered to amplify as its to_amplify argument.
    amplified_greet_guests_chain = InjectArgs(**self.abog) | greet_guests.s() | amplify.s(to_amplify='@guests_greeting')

    # Chains can be enqueued just like signatures. You can consider a signature a chain with only one service.
    chain_results = self.enqueue_child_and_get_results(amplified_greet_guests_chain)
    return chain_results['amplified_message']


@app.task()
@returns('job_title')
def get_springfield_power_plant_job_title(name):
    username_to_title = {'Charles Montgomery Burns': 'OWNER',
                         'Waylon Smithers': 'EXECUTIVE ASSISTANT',
                         'Lenny Leonard': 'DIRECTOR',
                         'Homer Simpson': 'SUPERVISOR'}
    return username_to_title.get(name, 'UNKNOWN')


@InputConverter.register
@SingleArgDecorator('employee_names')
def employee_names_to_list(employee_names):
    return employee_names.split(',')


@app.task(bind=True, returns=['amplified_greeting'])
def greet_springfield_power_plant_employees(self, employee_names):
    names_with_titles = []
    for name in employee_names:
        job_title = self.enqueue_child_and_get_results(get_springfield_power_plant_job_title.s(name=name))['job_title']
        names_with_titles.append(f"{job_title} {name}")

    results = self.enqueue_child_and_get_results(amplified_greet_guests.s(guests=names_with_titles))
    return results['amplified_greeting']


# A contrived example showing how CopyBogKeys can be used to copy values in between chained services, preventing
# data from being trampled when multiple services within a chain return values with the same name.
# Of course, in practice, it's preferable to enqueue the multiple calls to greet separately.
@app.task(bind=True, returns=['lee_greeting', 'tom_greeting'])
def greet_lee_and_tom(self: FireXTask):
    chain = (greet.s("Lee")
             | CopyBogKeys.s({'greeting': 'lee_greeting'})
             | greet.s("Tom"))
    results = self.enqueue_child_and_get_results(chain)
    # Lee's greeting has been copied to 'lee_greeting', so it's still accessible after Tom's greet has executed.
    lee_greeting = results['lee_greeting']
    # Tom's greeting still has the return name specified in the greet service.
    tom_greeting = results['greeting']
    return lee_greeting, tom_greeting



"""
Aggregates events in to the task data model.
"""
from collections import namedtuple
from datetime import datetime
import logging

from firexapp.events.model import ALL_RUNSTATES, INCOMPLETE_RUNSTATES, COMPLETE_RUNSTATES, TaskColumn

logger = logging.getLogger(__name__)


EventAggregatorConfig = namedtuple('EventAggregatorConfig',
                                   ['copy_fields', 'merge_fields', 'keep_initial_fields', 'field_to_celery_transforms'])


#
# config field options:
#   copy_celery - True if this field should be copied from the celery event to the task data model. If the field already
#                   has a value on the data model, more recent celery field values will overwrite existing values by
#                   default. If overwriting should be avoided, see 'aggregate_merge' and 'aggregate_keep_initial'
#                   options described below.
#
#   transform_celery - A function to be executed on the entire event when the corresponding key is present in a celery
#                       event. The function returns a dict that dict.update the existing data from the event, possibly
#                       overwriting data copied from the celery event by copy_celery=True. Can be used to change
#                       the field name on the data model from the field name from celery.
#
#   aggregate_merge - True if model updates should deep merge collection data types (lists, dicts, sets) instead of
#                       overwriting.
#
#   aggregate_keep_initial - True if data model field updates should be ignored after an initial value has been set.
#                               This is one way of preventing overwriting, see also 'aggregate_merge'.
#
#
FIELD_CONFIG = {
    TaskColumn.UUID.value: {'copy_celery': True},
    TaskColumn.HOSTNAME.value: {'copy_celery': True},
    TaskColumn.PARENT_ID.value: {'copy_celery': True},
    'type': {
        'copy_celery': True,
        'transform_celery': lambda e: {
            'state': e['type'],
            'states': [{TaskColumn.STATE.value: e['type'],
                        'timestamp': e.get('timestamp', None)}],
        } if e['type'] in ALL_RUNSTATES else {},
    },
    TaskColumn.RETRIES.value: {'copy_celery': True},
    TaskColumn.BOUND_ARGS.value: {'copy_celery': True},
    TaskColumn.ACTUAL_RUNTIME.value: {'copy_celery': True},
    TaskColumn.UTCOFFSET.value: {'copy_celery': True},
    TaskColumn.DEFAULT_BOUND_ARGS.value: {'copy_celery': True},
    TaskColumn.FROM_PLUGIN.value: {'copy_celery': True},
    TaskColumn.RESULTS.value: {'copy_celery': True},
    TaskColumn.TRACEBACK.value: {'copy_celery': True},
    TaskColumn.EXCEPTION.value: {'copy_celery': True},
    TaskColumn.LONG_NAME.value: {
        'copy_celery': True,
        'transform_celery': lambda e: {'name': e['long_name'].split('.')[-1]},
    },
    TaskColumn.NAME.value: {
        # TODO: firexapp should send long_name, since it will overwrite 'name' copied from celery. Then get rid of
        # the following config.
        'transform_celery': lambda e: {TaskColumn.NAME.value: e[TaskColumn.NAME.value].split('.')[-1],
                                       TaskColumn.LONG_NAME.value: e[TaskColumn.NAME.value]},
    },
    TaskColumn.CHAIN_DEPTH.value: {'copy_celery': True},
    TaskColumn.FIRST_STARTED.value: {'aggregate_keep_initial': True},
    TaskColumn.EXCEPTION_CAUSE_UUID.value: {'copy_celery': True},
    'states': {'aggregate_merge': True},
    'url': {
        # TODO: only for backwards compat. Can use log_filepath.
        'transform_celery': lambda e: {TaskColumn.LOGS_URL.value: e['url']},
    },
    'log_filepath': {
        'transform_celery': lambda e: {TaskColumn.LOGS_URL.value: e['log_filepath']},
    },
    'local_received': {
        # Note first_started is never overwritten by aggregation.
        'transform_celery': lambda e: {TaskColumn.FIRST_STARTED.value: e['local_received']},
    },
}


def _get_keys_with_true(input_dict, key):
    return [k for k, v in input_dict.items() if v.get(key, False)]


DEFAULT_AGGREGATOR_CONFIG = EventAggregatorConfig(
    copy_fields=_get_keys_with_true(FIELD_CONFIG, 'copy_celery'),
    merge_fields=_get_keys_with_true(FIELD_CONFIG, 'aggregate_merge'),
    keep_initial_fields=_get_keys_with_true(FIELD_CONFIG, 'aggregate_keep_initial'),
    field_to_celery_transforms={k: v['transform_celery'] for k, v in FIELD_CONFIG.items() if 'transform_celery' in v},
)


def _deep_merge_keys(dict1, dict2, keys):
    dict1_to_merge = {k: v for k, v in dict1.items() if k in keys}
    dict2_to_merge = {k: v for k, v in dict2.items() if k in keys}
    return _deep_merge(dict1_to_merge, dict2_to_merge)


def _both_instance(o1, o2, _type):
    return isinstance(o1, _type) and isinstance(o2, _type)


def _deep_merge(dict1, dict2):
    result = dict(dict1)
    for d2_key in dict2:
        if d2_key in dict1:
            v1 = dict1[d2_key]
            v2 = dict2[d2_key]
            if _both_instance(v1, v2, dict):
                result[d2_key] = _deep_merge(v1, v2)
            elif _both_instance(v1, v2, list):
                result[d2_key] = v1 + v2
            elif _both_instance(v1, v2, set):
                result[d2_key] = v1.union(v2)
            elif v1 == v2:
                # already the same value in both dicts, take from either.
                result[d2_key] = v1
            else:
                # Both d1 and d2 have entries for d2_key, both entries are not dicts or lists or sets,
                # and the values are not the same. This is a conflict.
                # Overwrite d1's value to simulate dict.update() behaviour.
                result[d2_key] = v2
        else:
            # New key for d1, just add it.
            result[d2_key] = dict2[d2_key]
    return result


# Event data extraction/transformation without current state context.
def get_new_event_data(event, copy_fields, field_to_celery_transforms):
    new_task_data = {}
    for field in copy_fields:
        if field in event:
            new_task_data[field] = event[field]

    # Note if a field is both a copy field and a transform, the transform overrides if the output writes to the same
    # key.
    for field, transform in field_to_celery_transforms.items():
        if field in event:
            new_task_data.update(transform(event))

    return {event['uuid']: new_task_data}


def find_data_changes(task, new_task_data, keep_initial_fields, merge_fields):
    # Some fields overwrite whatever is present. Be permissive, since not all fields captured are from celery,
    # so not all have entries in the field config.
    no_overwrite_fields = keep_initial_fields + merge_fields
    override_dict = {k: v for k, v in new_task_data.items() if k not in no_overwrite_fields}

    changed_data = {}
    for new_data_key, new_data_val in override_dict.items():
        if new_data_key not in task or task[new_data_key] != new_data_val:
            changed_data[new_data_key] = new_data_val

    # Some field updates are dropped if there is already a value for that field name (keep initial).
    for no_overwrite_key in keep_initial_fields:
        if no_overwrite_key in new_task_data and no_overwrite_key not in task:
            changed_data[no_overwrite_key] = new_task_data[no_overwrite_key]

    # Some fields need to be accumulated across events, not overwritten from latest event.
    merged_values = _deep_merge_keys(task, new_task_data, merge_fields)
    for merged_data_key, merged_data_val in merged_values.items():
        if merged_data_key not in task or task[merged_data_key] != merged_data_val:
            changed_data[merged_data_key] = merged_data_val

    return changed_data


class FireXEventAggregator:
    """ Aggregates many events in to the task data model. """

    def __init__(self, aggregator_config: EventAggregatorConfig = DEFAULT_AGGREGATOR_CONFIG):
        self.tasks_by_uuid = {}
        self.new_task_num = 1
        self.root_uuid = None
        self.aggregator_config = aggregator_config

    def aggregate_events(self, events):
        new_data_by_task_uuid = {}
        for e in events:
            event_new_data_by_task_uuid = self._aggregate_event(e)
            for uuid, new_data in event_new_data_by_task_uuid.items():
                if uuid not in new_data_by_task_uuid:
                    new_data_by_task_uuid[uuid] = {}
                new_data_by_task_uuid[uuid].update(new_data)
        return new_data_by_task_uuid

    def generate_incomplete_events(self):
        """
        Unfortunately, if a run terminates ungracefully, incomplete tasks will never arrive at a
        terminal runstate. The 'task-incomplete' runstate is a fake (non-backend) terminal runstate
        that is generated here so that the UI can show a non-incomplete runstate.
        :return:
        """
        return [{'uuid': task['uuid'], 'type': 'task-incomplete', 'timestamp': datetime.now().timestamp()}
                for task in self.tasks_by_uuid.values()
                if task['state'] in INCOMPLETE_RUNSTATES]

    def is_root_complete(self):
        if not self.root_uuid or self.root_uuid not in self.tasks_by_uuid:
            return False  # Might not have root event yet.
        root_runstate = self.tasks_by_uuid[self.root_uuid].get('state', None)
        return root_runstate in COMPLETE_RUNSTATES

    def _get_or_create_task(self, task_uuid):
        if task_uuid not in self.tasks_by_uuid:
            task = {
                'uuid': task_uuid,
                'task_num': self.new_task_num,
            }
            self.new_task_num += 1
            self.tasks_by_uuid[task_uuid] = task
            is_new = True
        else:
            task = self.tasks_by_uuid[task_uuid]
            is_new = False
        return task, is_new

    def _aggregate_event(self, event):
        if ('uuid' not in event
                # The uuid can be null, it's unclear what this means but it can't be associated with a task.
                or not event['uuid']
                # Revoked events can be sent before any other, and we'll never get any data (name, etc) for that task.
                # Therefore ignore events that are for a new UUID that have revoked type.
                or (event['uuid'] not in self.tasks_by_uuid and event.get('type', '') == 'task-revoked')):
            return {}

        if event.get(TaskColumn.PARENT_ID.value, '__no_match') is None and self.root_uuid is None:
            self.root_uuid = event['uuid']

        new_data_by_task_uuid = get_new_event_data(event,
                                                   self.aggregator_config.copy_fields,
                                                   self.aggregator_config.field_to_celery_transforms)
        changes_by_task_uuid = {}
        for task_uuid, new_task_data in new_data_by_task_uuid.items():
            task, is_new_task = self._get_or_create_task(task_uuid)

            changed_data = find_data_changes(task,
                                             new_task_data,
                                             self.aggregator_config.keep_initial_fields,
                                             self.aggregator_config.merge_fields)
            task.update(changed_data)

            # If we just created the task, we need to send the auto-initialized fields, as well as data from the event.
            # If this isn't a new event, we only need to send what has changed.
            changes_by_task_uuid[task_uuid] = dict(task) if is_new_task else dict(changed_data)

        return changes_by_task_uuid



import json
import os
from socket import gethostname

from firexapp.application import get_app_tasks
from firexapp.common import silent_mkdir, create_link
from firexapp.submit.reporting import ReportGenerator
from firexkit.result import get_results
from firexkit.task import convert_to_serializable
from celery.utils.log import get_task_logger
from firexapp.engine.celery import app

logger = get_task_logger(__name__)


class FireXJsonReportGenerator(ReportGenerator):
    formatters = ('json',)

    reporter_dirname = 'json_reporter'
    initial_report_filename = 'initial_report.json'
    completion_report_filename = 'completion_report.json'
    report_link_filename = 'run.json'

    @staticmethod
    def get_common_run_data(uid, chain, submission_dir, argv, original_cli):
        data = {'chain': [t.short_name for t in get_app_tasks(chain)],
                'firex_id': str(uid),
                'logs_path': uid.logs_dir,
                'submission_host': app.conf.mc or gethostname(),
                'submission_dir': submission_dir,
                'submission_cmd': original_cli or list(argv),
                }

        viewers = uid.viewers
        if viewers:
            data.update(viewers)

        return data

    @staticmethod
    def write_report_file(data, report_file):
        # Create the json_reporter dir if it doesn't exist
        silent_mkdir(os.path.dirname(report_file))

        with open(report_file, 'w') as f:
            json.dump(convert_to_serializable(data),
                      fp=f,
                      skipkeys=True,
                      sort_keys=True,
                      indent=4)

    def pre_run_report(self, uid, chain, submission_dir, argv, original_cli=None, json_file=None, **kwargs):
        data = self.get_common_run_data(uid=uid, chain=chain, submission_dir=submission_dir, argv=argv,
                                        original_cli=original_cli)
        data['completed'] = False
        report_file = os.path.join(uid.logs_dir, self.reporter_dirname, self.initial_report_filename)
        self.write_report_file(data, report_file)

        report_link = os.path.join(uid.logs_dir, self.report_link_filename)
        try:
            create_link(report_file, report_link, delete_link=False)
        except FileExistsError:
            logger.debug(f'f{report_link} link already exist; post_run must have already run. '
                         f'No need to link to f{report_file}')

        if json_file:
            try:
                create_link(report_link, json_file, delete_link=False, relative=True)
            except FileExistsError:
                logger.debug(f'{json_file} link already exist; '
                             f'post_run must have already created the link to {report_link}')

    def post_run_report(self, uid, chain, root_id, submission_dir, argv, original_cli=None, json_file=None, **kwargs):
        data = self.get_common_run_data(uid=uid, chain=chain, submission_dir=submission_dir, argv=argv,
                                        original_cli=original_cli)
        data['completed'] = True
        data['results'] = get_results(root_id)
        report_file = os.path.join(uid.logs_dir, self.reporter_dirname, self.completion_report_filename)

        self.write_report_file(data, report_file)

        report_link = os.path.join(uid.logs_dir, self.report_link_filename)
        create_link(report_file, report_link, relative=True)

        if json_file:
            try:
                # This is typically not required, unless post_run ran before pre_run
                create_link(report_link, json_file, delete_link=False, relative=True)
            except FileExistsError:
                # This is expected for most cases
                pass

    def add_entry(self, key_name, value, priority, formatters, **extra):
        pass


def get_completion_report_data(logs_dir):
    report_file = os.path.join(logs_dir, FireXJsonReportGenerator.reporter_dirname,
                               FireXJsonReportGenerator.completion_report_filename)
    with open(report_file) as f:
        return json.load(fp=f)



import sys
import time
import os
import inspect
import subprocess

from firexkit.resources import get_cloud_ci_install_config_path
from firexapp.submit.submit import get_firex_id_from_output, get_log_dir_from_output
from firexapp.submit.tracking_service import has_flame
from firexapp.submit.install_configs import load_new_install_configs, INSTALL_CONFIGS_ENV_NAME
from firexapp.testing.config_base import InterceptFlowTestConfiguration, FlowTestConfiguration


class ConfigInterpreter:
    execution_directory = None

    def __init__(self):
        self.profile = False
        self.coverage = False
        self.is_public = False

    @staticmethod
    def is_submit_command(test_config: FlowTestConfiguration):
        cmd = test_config.initial_firex_options()
        if not cmd:
            return False
        if cmd[0] in "submit":
            return True
        return cmd[0] not in ["list", "info"]

    @staticmethod
    def is_instance_of_intercept(test_config: FlowTestConfiguration):
        base_classes = [cls.__name__ for cls in inspect.getmro(type(test_config))[1:]]
        return InterceptFlowTestConfiguration.__name__ in base_classes

    @staticmethod
    def create_mock_file(results_folder, results_file, test_name, intercept_microservice):
        mock_file_name = test_name + "_mock.py"
        mock_file = os.path.join(results_folder, mock_file_name)
        intercept_task = """
import os
from celery import current_app as app
from firexkit.task import FireXTask


# noinspection PyPep8Naming
@app.task(base=FireXTask)
def {0}(**kwargs):
    local_stuff = locals()
    local_stuff.update(kwargs)
    import pickle

    def str_default(o):
        return str(o)
    from helper import str2file
    file_path = "{1}"
    dir_path = os.path.dirname(file_path)
    if not os.path.isdir(dir_path):
        os.makedirs(dir_path)
    with open(file_path, "wb") as f:
        pickle.dump(local_stuff, f)
    if not os.path.isfile(file_path):
        raise Exception("Could not create result files")
    """
        content = intercept_task.format(intercept_microservice, results_file)
        if not os.path.isdir(results_folder):
            os.mkdir(results_folder)
        with open(mock_file, "w") as f:
            f.write(content)
        return mock_file

    def document_viewer(self, file_path: str)->str:
        return file_path

    @staticmethod
    def get_intercept_results_file(flow_test_config):
        return os.path.join(flow_test_config.results_folder, flow_test_config.name + ".results")

    @staticmethod
    def get_test_name(flow_test_config):
        return flow_test_config.__class__.__name__

    def run_integration_test(self, flow_test_config, results_folder):
        # provide sub-folder for testsuite data
        flow_test_config.results_folder = os.path.join(results_folder, flow_test_config.name)
        os.makedirs(flow_test_config.results_folder)
        flow_test_config.std_out = os.path.join(flow_test_config.results_folder, flow_test_config.name + ".stdout.txt")
        flow_test_config.std_err = os.path.join(flow_test_config.results_folder, flow_test_config.name + ".stderr.txt")

        cmd = self.create_cmd(flow_test_config)
        self.run_executable(cmd, flow_test_config)

    def create_cmd(self, flow_test_config)->[]:
        # assemble options, adding/consolidating --external and --sync
        cmd = self.get_exe(flow_test_config)
        cmd += flow_test_config.initial_firex_options()

        plugins = self.collect_plugins(flow_test_config)
        if plugins:
            cmd += ["--plugins", ",".join(plugins)]

        submit_test = self.is_submit_command(flow_test_config)
        if submit_test:
            flow_test_config.logs_link = os.path.join(flow_test_config.results_folder, flow_test_config.name + ".logs")
            cmd += ["--logs_link", flow_test_config.logs_link]
            if getattr(flow_test_config, "sync", True):
                cmd += ["--sync"]
            if has_flame() and getattr(flow_test_config, "flame_terminate_on_complete", True):
                cmd += ["--flame_terminate_on_complete"]
            if (self.is_public # only use public CI install config for is_public runs
                    and not getattr(flow_test_config, 'no_install_config', False)
                    # Some tests supply install configs (via CLI or ENV) for legitimate testing purposes.
                    and '--install_configs' not in cmd
                    and INSTALL_CONFIGS_ENV_NAME not in os.environ):
                # TODO: should merge test-specific install_configs with ci-viewer configs,
                #  since we usually want the ci URLs, even with a test's install_config specifies other stuff.
                cmd += ['--install_configs', get_cloud_ci_install_config_path()]

        return cmd

    def get_exe(self, flow_test_config)->[]:
        import firexapp
        if self.coverage and not hasattr(flow_test_config, 'no_coverage'):
            return ["coverage", "run", "--branch", "--append", "-m", firexapp.__name__]
        if self.profile:
            base_dir = os.path.dirname(firexapp.__file__)
            exe_file = os.path.join(base_dir, "__main__.py")
            return ["python", "-m", "cProfile", "-s", "cumtime", exe_file]
        return ["python", "-m", firexapp.__name__]

    def collect_plugins(self, flow_test_config)->[]:
        # add test file and dynamically generated files to --plugins
        test_src_file = inspect.getfile(flow_test_config.__class__)
        plugins = [
            os.path.realpath(test_src_file),  # must be first to be imported
        ]

        submit_test = self.is_submit_command(flow_test_config)
        if submit_test:
            if self.coverage and not hasattr(flow_test_config, 'no_coverage'):
                # add the coverage plugin to restart celery in coverage mode
                from firexapp.testing import coverage_plugin
                plugins.append(coverage_plugin.__file__)

            if self.is_instance_of_intercept(flow_test_config):
                # create file containing mock and capture microservices
                intercept = flow_test_config.intercept_service()
                if intercept:
                    intercept_results_file = self.get_intercept_results_file(flow_test_config)
                    mock_file = self.create_mock_file(flow_test_config.results_folder,
                                                      intercept_results_file,
                                                      flow_test_config.name, intercept)
                    plugins.append(mock_file)

        # test and use add_plugins=False to explicitly prevent dynamic plugins being added to the test
        extras = getattr(flow_test_config, "add_plugins", [])
        if extras:
            plugins += extras
        if extras is False:
            plugins = []
        return plugins

    def run_executable(self, cmd, flow_test_config):

        # print useful links
        test_src_file = inspect.getfile(flow_test_config.__class__)
        if hasattr(flow_test_config, 'logs_link'):
            print("\tLogs:", self.document_viewer(flow_test_config.logs_link), file=sys.stderr)
        print("\tTest source:", self.document_viewer(test_src_file), file=sys.stderr)
        print("\tStdout:", self.document_viewer(flow_test_config.std_out), file=sys.stderr)
        print("\tStderr:", self.document_viewer(flow_test_config.std_err), file=sys.stderr)


        elapsed_time = None
        verification_time = None

        # run firex
        try:
            with open(flow_test_config.std_out, 'w') as std_out_f, open(flow_test_config.std_err, 'w') as std_err_f:
                start_time = time.monotonic()
                process = subprocess.Popen(cmd, stdout=std_out_f, stderr=std_err_f,
                                           universal_newlines=True, shell=False, cwd=self.execution_directory)
                _, _ = process.communicate(timeout=getattr(flow_test_config, "timeout", None))
                elapsed_time = time.monotonic() - start_time

            start_time = time.monotonic()
            # check for expected return code
            expected_return = flow_test_config.assert_expected_return_code(process.returncode)
            if expected_return is not None:
                raise Exception("assert_expected_return_code should not return. It should assert if needed")

            if self.is_submit_command(flow_test_config) and process.returncode == 0 and \
                    self.is_instance_of_intercept(flow_test_config) and \
                    flow_test_config.intercept_service():
                # retrieve captured kwargs and validate them
                intercept_results_file = self.get_intercept_results_file(flow_test_config)
                if not os.path.isfile(intercept_results_file):
                    raise FileNotFoundError(intercept_results_file + " was not found. Could not retrieve results.")

                import pickle
                with open(intercept_results_file, 'rb') as results_file_f:
                    captured_options = pickle.load(results_file_f)
                flow_test_config.assert_expected_options(captured_options)

            with open(flow_test_config.std_out, 'r') as std_out_f, open(flow_test_config.std_err, 'r') as std_err_f:
                errors = std_err_f.read().split("\n")
                errors = [line for line in errors if line and not line.startswith("pydev debugger:")]
                flow_test_config.assert_expected_firex_output(std_out_f.read(), "\n".join(errors))
            verification_time = time.monotonic() - start_time
        except (subprocess.TimeoutExpired, KeyboardInterrupt) as e:
            print("\t%s!" % type(e).__name__, file=sys.stderr)
            self.cleanup_after_timeout(flow_test_config.std_out, flow_test_config.std_err)
            raise
        except Exception as e:
            print('\tException: {}: {}'.format(type(e).__name__, e), file=sys.stderr)
            raise
        finally:
            self.on_test_exit(flow_test_config.std_out, flow_test_config.std_err)

            # report on time
            if elapsed_time is not None:
                msg = "\tTime %.1fs" % elapsed_time
                if verification_time is not None:
                    overhead = " +(%.1fs)" % verification_time
                    if overhead != " +(0.0s)":
                        msg += overhead
                print(msg, file=sys.stderr)

    def cleanup_after_timeout(self, std_out, std_err):
        pass

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def on_test_exit(self, std_out, std_err):
        # Try to print to log directory to help with debugging
        # noinspection PyBroadException
        try:
            with open(std_out, 'r') as std_out_f:
                std_out_content = std_out_f.read()
                firex_id = get_firex_id_from_output(std_out_content)
                if firex_id:
                    print("\tFireX ID: " + firex_id, file=sys.stderr)
                    if self.is_public:
                        install_configs = load_new_install_configs(firex_id,
                                                                   get_log_dir_from_output(std_out_content),
                                                                   get_cloud_ci_install_config_path())
                        print(f'\tLogs URL: {install_configs.get_logs_root_url()}', file=sys.stderr)
                        print(f"\tFlame: {install_configs.run_url}", file=sys.stderr)
        except Exception as e:
            print(e)
            pass



import os
import signal
import tempfile
from argparse import ArgumentParser, RawTextHelpFormatter

from firexapp.plugins import load_plugin_modules, cdl2list
from firexapp.submit.console import setup_console_logging
from firexkit.permissions import DEFAULT_UMASK

logger = setup_console_logging(__name__)


def main():
    os.umask(DEFAULT_UMASK)
    # Need to call setup_console_logging like this as this module is always called from another.
    setup_console_logging("__main__")
    with tempfile.NamedTemporaryFile(delete=True) as submission_tmp_file:
        from firexapp.submit.submit import SubmitBaseApp
        submit_app = SubmitBaseApp(submission_tmp_file=submission_tmp_file.name)
        app = FireXBaseApp(submit_app=submit_app)
        app.run()


def import_microservices(plugins_files=None, imports: tuple = None) -> []:
    for f in cdl2list(plugins_files):
        if not os.path.isfile(f):
            raise FileNotFoundError(f)

    from firexapp.engine.celery import app

    if not imports:
        imports = app.conf.imports

    for module_name in imports:
        __import__(module_name)

    load_plugin_modules(plugins_files)

    return app.tasks


def get_app_task(task_short_name: str, all_tasks=None):
    task_short_name = task_short_name.strip()
    if all_tasks is None:
        from firexapp.engine.celery import app
        all_tasks = app.tasks

    # maybe it isn't a short name, but a long one
    if task_short_name in all_tasks:
        return all_tasks[task_short_name]

    # Search for an exact match first
    for key, value in all_tasks.items():
        if key.split('.')[-1] == task_short_name:
            return value

    # Let's do a case-insensitive search
    task_name_lower = task_short_name.lower()
    for key, value in all_tasks.items():
        if key.split('.')[-1].lower() == task_name_lower:
            return value

    # Can't find a match
    from celery.exceptions import NotRegistered
    raise NotRegistered(task_short_name)


def get_app_tasks(tasks, all_tasks=None):
    if type(tasks) is str:
        tasks = tasks.split(",")
    return [get_app_task(task, all_tasks) for task in tasks]


class FireXBaseApp:
    def __init__(self, submit_app=None, info_app=None):
        if not info_app:
            from firexapp.info import InfoBaseApp
            info_app = InfoBaseApp()
        self.info_app = info_app

        if not submit_app:
            from firexapp.submit.submit import SubmitBaseApp
            submit_app = SubmitBaseApp()
        self.submit_app = submit_app
        self.arg_parser = None
        self.running_app = None
        self._signal_exit_handler = ExitSignalHandler(self)
        self.submit_args_to_process = None

    def run(self, sys_argv=None):
        if not self.arg_parser:
            self.arg_parser = self.create_arg_parser()

        try:
            if sys_argv is not None:
                "".join(sys_argv).encode('ascii')
        except UnicodeEncodeError as ue:
            self.arg_parser.error(
                'You entered a non-ascii character at the command line.\n' + str(ue))

        arguments, others = self.arg_parser.parse_known_args(sys_argv)

        # run default help
        if not hasattr(arguments, "func"):
            self.arg_parser.print_help()
            self.arg_parser.exit()
        if self.submit_app.run_submit.__name__ not in arguments.func.__name__:
            if len(others):
                # only submit supports 'other' arguments
                msg = 'Unrecognized arguments: %s' % ' '.join(others)
                self.arg_parser.error(message=msg)
            arguments.func(arguments)
        else:
            self.running_app = self.submit_app
            self.submit_app.submit_args_to_process = sys_argv
            arguments.func(arguments, others)

    def main_error_exit_handler(self, reason=None):
        if self.running_app and hasattr(self.running_app, self.main_error_exit_handler.__name__):
            self.running_app.main_error_exit_handler(reason=reason)
        exit(-1)

    def create_arg_parser(self, description=None)->ArgumentParser:
        if not description:
            description = """
FireX is a workflow automation and execution engine built using a micro-service oriented design and architecture.
FireX provides a framework to facilitate the automation of the various workflows that are part of every development
and testing processes."""
        main_parser = ArgumentParser(description=description, formatter_class=RawTextHelpFormatter)
        sub_parser = main_parser.add_subparsers()

        self.info_app.create_list_sub_parser(sub_parser)
        self.info_app.create_info_sub_parser(sub_parser)
        submit_parser = self.submit_app.create_submit_parser(sub_parser)

        self.arg_parser = main_parser

        self.submit_app.store_parser_attributes(main_parser, submit_parser)

        return main_parser


class ExitSignalHandler:
    first_warning = "\nExiting due to signal %s"
    second_warning = "\nWe know! Have a little patience for crying out loud!"
    last_warning = "\nFINE! We'll stop. But you might have leaked a celery instance or a broker instance."

    @staticmethod
    def _register_signal_handlers(handler):
        signal.signal(signal.SIGTERM, handler)
        signal.signal(signal.SIGINT, handler)
        signal.signal(signal.SIGHUP, handler)

    def __init__(self, app):
        def first_exit_handler(signal_num, _):
            def last_exit_handler(_, __):
                logger.error(self.last_warning)
                exit(-1)

            def second_exit_handler(_, __):
                logger.error(self.second_warning)
                self._register_signal_handlers(last_exit_handler)

            self._register_signal_handlers(second_exit_handler)
            signal_name = signal.Signals(signal_num).name
            logger.error(self.first_warning % signal_name)
            app.main_error_exit_handler(reason=f"Received signal {signal_name}.")

        self._register_signal_handlers(first_exit_handler)



import json
import os
import secrets
import time
import shlex
import subprocess
from functools import partial

from firexapp.fileregistry import FileRegistry
from firexapp.submit.uid import Uid
from socket import gethostname
from urllib.parse import urlsplit
from logging import ERROR, INFO
from psutil import Process
from pathlib import Path

from firexapp.broker_manager import BrokerManager
from firexapp.common import get_available_port, wait_until, silent_mkdir

REDIS_DIR_REGISTRY_KEY = 'REDIS_DIR_REGISTRY_KEY'
FileRegistry().register_file(REDIS_DIR_REGISTRY_KEY, os.path.join(Uid.debug_dirname, 'redis'))

REDIS_LOG_REGISTRY_KEY = 'REDIS_LOG_REGISTRY_KEY'
FileRegistry().register_file(REDIS_LOG_REGISTRY_KEY,
                             os.path.join(FileRegistry().get_relative_path(REDIS_DIR_REGISTRY_KEY), 'redis.stdout.txt'))

REDIS_PID_REGISTRY_KEY = 'REDIS_PID_REGISTRY_KEY'
FileRegistry().register_file(REDIS_PID_REGISTRY_KEY,
                             os.path.join(FileRegistry().get_relative_path(REDIS_DIR_REGISTRY_KEY), 'redis.pid'))

REDIS_METADATA_REGISTRY_KEY = 'REDIS_METADATA_REGISTRY_KEY'
FileRegistry().register_file(REDIS_METADATA_REGISTRY_KEY,
                             os.path.join(FileRegistry().get_relative_path(REDIS_DIR_REGISTRY_KEY),
                                          'run-metadata.json'))

REDIS_CREDS_REGISTRY_KEY = 'REDIS_CREDS_REGISTRY_KEY'
FileRegistry().register_file(REDIS_CREDS_REGISTRY_KEY,
                             os.path.join(FileRegistry().get_relative_path(REDIS_DIR_REGISTRY_KEY),
                                          'run-credentials.json'))


class RedisDidNotBecomeActive(Exception):
    pass


class RedisPortNotAssigned(Exception):
    pass


class RedisPasswordReadError(Exception):
    pass


class RedisManager(BrokerManager):

    _METADATA_BROKER_URL_KEY = 'broker_url'  # For reading old-style configurations, before the usage of passwords
    _METADATA_BROKER_HOST_KEY = 'broker_host'
    _METADATA_BROKER_PORT_KEY = 'broker_port'
    _BROKER_PASSWORD_KEY = 'broker_password'
    _BROKER_FAILED_AUTH_STR = 'NOAUTH Authentication required'  # Actual output from Redis

    def __init__(self, redis_bin_base, hostname=gethostname(), port=None, logs_dir=None, password=None):
        self.redis_bin_base = redis_bin_base
        self.host = hostname
        self.port = port
        self.logs_dir = logs_dir

        self._password = str(password) if password else secrets.token_urlsafe(32).lstrip('-')
        self._log_file = None
        self._pid_file = None
        self._metadata_file = None
        self._password_file = None
        self._redis_server_bin = os.path.join(redis_bin_base, 'redis-server')

        if logs_dir:
            # Run Redis server from logs directory, so that the command line has the logs path
            redis_bin_dir = os.path.join(logs_dir, 'bin')
            redis_server_bin = os.path.join(redis_bin_dir, os.path.basename(self._redis_server_bin))
            try:
                silent_mkdir(redis_bin_dir)
                os.symlink(self._redis_server_bin, redis_server_bin)
            except FileExistsError:
                # Link was created previously
                pass
            self._redis_server_bin = redis_server_bin

    @property
    def redis_cli_cmd(self):
        return self.get_redis_cli_cmd(self.port)

    def get_redis_cli_cmd(self, port, include_host=False):
        cmd = os.path.join(self.redis_bin_base, 'redis-cli') + ' -p %d -a %s' % (port, self._password)
        if include_host or self.host != gethostname():
            cmd += ' -h %s' % self.host
        return cmd

    @property
    def redis_server_cmd(self):
        return self.get_redis_server_cmd(self.port)

    def get_redis_server_cmd(self, port):
        return self._redis_server_bin + ' --port %d --requirepass %s' \
               % (port, self._password)

    @property
    def broker_url(self):
        return self.get_broker_url(self.port, self.host, self._password)

    @property
    def broker_url_safe_print(self):
        return self.get_broker_url(self.port, self.host, '**')

    @property
    def port(self):
        if not self.__port:
            raise RedisPortNotAssigned()
        return self.__port

    @port.setter
    def port(self, port):
        self.__port = int(port) if port else port

    @staticmethod
    def get_log_file(logs_dir):
        return FileRegistry().get_file(REDIS_LOG_REGISTRY_KEY, logs_dir)

    @staticmethod
    def get_pid_file(logs_dir):
        return FileRegistry().get_file(REDIS_PID_REGISTRY_KEY, logs_dir)

    @staticmethod
    def get_metadata_file(logs_dir):
        return FileRegistry().get_file(REDIS_METADATA_REGISTRY_KEY, logs_dir)

    @staticmethod
    def get_password_file(logs_dir):
        return FileRegistry().get_file(REDIS_CREDS_REGISTRY_KEY, logs_dir)

    @classmethod
    def read_metadata(cls, logs_dir):
        with open(cls.get_metadata_file(logs_dir)) as fp:
            metadata = json.load(fp)
        return metadata

    @classmethod
    def read_password_data(cls, logs_dir):
        try:
            with open(cls.get_password_file(logs_dir)) as fp:
                password_data = json.load(fp)
        except (PermissionError, FileNotFoundError) as e:
            raise RedisPasswordReadError('Cannot get broker password') from e

        return password_data

    @classmethod
    def get_hostname_port_from_logs_dir(cls, logs_dir):
        metadata = cls.read_metadata(logs_dir)
        try:
            return metadata[cls._METADATA_BROKER_HOST_KEY], str(metadata[cls._METADATA_BROKER_PORT_KEY])

        except KeyError as e:
            # Possibly old-style metadata, before broker password usage
            cls.log('Could not get hostname and port directly. Trying old method.', exc_info=e)
            return cls.get_hostname_port_from_url(broker_url=metadata[cls._METADATA_BROKER_URL_KEY])

    @classmethod
    def get_password_from_logs_dir(cls, logs_dir):
        password_data = cls.read_password_data(logs_dir)
        return password_data[cls._BROKER_PASSWORD_KEY]

    @classmethod
    def get_broker_failed_auth_str(cls) -> str:
        return cls._BROKER_FAILED_AUTH_STR

    @classmethod
    def get_broker_url_from_logs_dir(cls, logs_dir):
        hostname, port = cls.get_hostname_port_from_logs_dir(logs_dir)
        password = cls.get_password_from_logs_dir(logs_dir)

        return cls.get_broker_url(port, hostname, password)

    @property
    def log_file(self):
        if not self._log_file and self.logs_dir:
            _log_file = self.get_log_file(self.logs_dir)
            os.makedirs(os.path.dirname(_log_file), exist_ok=True)
            self._log_file = _log_file
        return self._log_file

    @property
    def pid_file(self):
        if not self._pid_file and self.logs_dir:
            _pid_file = self.get_pid_file(self.logs_dir)
            os.makedirs(os.path.dirname(_pid_file), exist_ok=True)
            self._pid_file = _pid_file
        return self._pid_file

    @property
    def metadata_file(self):
        if not self._metadata_file and self.logs_dir:
            _metadata_file = self.get_metadata_file(self.logs_dir)
            os.makedirs(os.path.dirname(_metadata_file), exist_ok=True)
            self._metadata_file = _metadata_file
        return self._metadata_file

    @property
    def password_file(self):
        if not self._password_file and self.logs_dir:
            _password_file = self.get_password_file(self.logs_dir)
            os.makedirs(os.path.dirname(_password_file), exist_ok=True)
            self._password_file = _password_file
        return self._password_file

    def create_metadata_file(self):
        if self.metadata_file:
            self.log('Creating %s' % self.metadata_file)
            data = {self._METADATA_BROKER_HOST_KEY: self.host, self._METADATA_BROKER_PORT_KEY: self.port}
            with open(self.metadata_file, 'w') as f:
                json.dump(data, f, sort_keys=True, indent=2)

    def create_password_file(self):
        if self.password_file:
            self.log('Creating %s' % self.password_file)
            data = {self._BROKER_PASSWORD_KEY: str(self._password)}
            # noinspection PyTypeChecker
            with open(self.password_file, 'w',  opener=partial(os.open, mode=0o600)) as f:
                json.dump(data, f, sort_keys=True, indent=2)

    def _start(self, timeout=60):
        try:
            port = self.port
        except RedisPortNotAssigned:
            port = get_available_port()
        self.log('Starting new process (port %d)...' % port)
        cmd = self.get_redis_server_cmd(port) + ' ' + '--loglevel debug ' \
                                                      '--protected-mode no ' \
                                                      '--daemonize yes ' \
                                                      '--timeout 0 ' \
                                                      '--client-output-buffer-limit slave 0 0 0 ' \
                                                      '--client-output-buffer-limit pubsub 0 0 0 ' \
                                                      '--save ""'
        if self.pid_file:
            cmd += ' --pidfile %s' % self.pid_file
        if self.log_file:
            cmd += ' --logfile %s' % self.log_file
        subprocess.check_call(shlex.split(cmd))
        if not wait_until(os.path.exists, timeout, 0.1, self.pid_file):
            raise RedisDidNotBecomeActive(f'The Redis pid file {self.pid_file} did not exist within {timeout}s')
        self.wait_until_active(port=port, timeout=timeout)
        self.port = port
        self.create_password_file()
        self.create_metadata_file()
        self.log('redis started.')

    def start(self, max_retries=3):
        max_trials = max_retries + 1
        trials = 0

        while True:
            trials += 1
            try:
                self._start()
            except (subprocess.CalledProcessError, RedisDidNotBecomeActive):
                if trials >= max_trials:
                    self.log('Redis did not come up after %d trial(s) (max_trials=%d)..Giving up!' %
                             (trials, max_trials), level=ERROR)
                    raise
                self.log('Redis did not come up after %d trial(s) (max_trials=%d)' % (trials, max_trials), level=INFO)
            else:
                break

    def shutdown(self):
        try:
            RedisManager.log('shutting down...')
            self.cli('shutdown')
        except subprocess.CalledProcessError:
            RedisManager.log('could not shutdown.')

    def force_kill(self):
        # noinspection PyBroadException
        try:
            RedisManager.log('force killing...')
            Process(int(Path(self.pid_file).read_text())).kill()
        except Exception:
            RedisManager.log('could not force kill.')

    def get_url(self) -> str:
        return self.broker_url

    def is_alive(self, port=None):
        port = self.port if port is None else port
        try:
            # including the host makes sure the current host can connect to itself via its hostname, like celery will.
            output = self.cli('PING', port=port, include_host=True)
        except subprocess.CalledProcessError:
            return False
        else:
            return output == 'PONG' or self.get_broker_failed_auth_str() in output

    def wait_until_active(self, timeout=60, port=None):
        port = self.port if port is None else port
        timeout_time = time.time() + timeout
        while time.time() < timeout_time:
            if self.is_alive(port=port):
                return
            time.sleep(0.1)
        raise RedisDidNotBecomeActive('Redis Server did not respond after %r seconds' % timeout)

    @staticmethod
    def get_broker_url(port=6379, hostname=gethostname(), password=None):
        preamble = f':{password}@' if password else ''
        return 'redis://%s%s:%d/0' % (preamble, hostname, int(port))

    @staticmethod
    def get_hostname_port_from_url(broker_url):
        url = urlsplit(broker_url)
        return url.hostname, str(url.port) if url.port else url.port

    @staticmethod
    def get_password_from_url(broker_url):
        return urlsplit(broker_url).password

    def get(self, key):
        return self.cli('GET %s' % key)

    def set(self, key, value):
        rc = self.cli('SET %s %s' % (key, value))
        assert rc == 'OK', 'The return value was %s' % rc

    def monitor(self, monitor_file):
        cmd = os.path.join(self.redis_bin_base, 'redis-cli') + ' -p %d -a %s MONITOR &' % (self.port, self._password)
        with open(monitor_file, 'w') as out:
            subprocess.check_call(cmd, shell=True, stdout=out, stderr=subprocess.STDOUT)

    def cli(self, cmd, port=None, include_host=False):
        port = self.port if port is None else port
        null = open(os.devnull, 'w')
        cmd = self.get_redis_cli_cmd(port=port, include_host=include_host) + ' %s' % cmd
        return subprocess.check_output(shlex.split(cmd), stderr=null).decode().strip()

    def __repr__(self):
        return self.broker_url

    def __eq__(self, other):
        return str(other) == self.broker_url



import logging

import celery.utils.log
from _socket import gethostname
from logging.handlers import WatchedFileHandler
import html
from celery.signals import after_setup_task_logger, after_setup_logger
import os
from firexapp.engine.celery import app
from firexkit.resources import get_firex_css_filepath, get_firex_logo_filepath
from firexkit.firexkit_common import JINJA_ENV
from celery._state import get_current_task

RAW_LEVEL_NAME = 'RAW'
PRINT_LEVEL_NAME = 'PRINT'


def add_hostname_to_log_records():
    old_factory = logging.getLogRecordFactory()

    def record_factory(*args, **kwargs):
        record = old_factory(*args, **kwargs)
        record.hostname = gethostname()
        return record

    logging.setLogRecordFactory(record_factory)


def log_raw(self, message, *args, **kwargs):
    self._log(logging.RAW, message, args, **kwargs)


def log_print(self, message, *args, **kwargs):
    self._log(logging.PRINT, message, args, **kwargs)


def add_custom_log_levels():
    logging.RAW = logging.DEBUG - 5
    logging.addLevelName(logging.RAW, RAW_LEVEL_NAME)
    logging.Logger.raw = log_raw

    logging.PRINT = logging.WARNING + 5
    logging.addLevelName(logging.PRINT, PRINT_LEVEL_NAME)
    logging.Logger.print = log_print

    # Need to add the PRINT custom level to kombu,
    # otherwise, the celery worker barfs at the bootup sequence
    add_print_custom_log_level_to_kombu()


def add_print_custom_log_level_to_kombu():
    from kombu.log import LOG_LEVELS
    LOG_LEVELS.setdefault(PRINT_LEVEL_NAME, logging.PRINT)
    LOG_LEVELS.setdefault(logging.PRINT, PRINT_LEVEL_NAME)
    LOG_LEVELS.setdefault(RAW_LEVEL_NAME, logging.RAW)
    LOG_LEVELS.setdefault(logging.RAW, RAW_LEVEL_NAME)


def html_escape(msg):
    msg = str(msg)
    try:
        return html.escape(msg)
    except TypeError:
        return html.escape(msg.decode('ascii', errors='ignore'))


class AddHtmlElementsToLogRecords(logging.Filter):
    def filter(self, record):
        # Add a span class element in the dict
        span_classes = record.levelname.lower()
        try:
            span_class = record.span_class
        except AttributeError:
            pass
        else:
            span_classes += f' {span_class}'
        record.span_class_element = f"<span class='{span_classes}'>"

        # Add a label element if it exists
        try:
            record.label_element += f'<a name="{record.label}"></a>'
        except AttributeError:
            record.label_element = ''

        return True

class FireXFormatter(celery.utils.log.ColorFormatter):
    def __init__(self, fmt):
        new_fmt = '%(span_class_element)s%(label_element)s' + fmt + '</span>'
        super().__init__(fmt=new_fmt, use_color=False)
        self.datefmt = '%m-%d %H:%M:%S %z'

    def format(self, record):
        if record.levelno == logging.RAW:
            original_format = self._style._fmt
            self._style._fmt = '%(message)s'
            mssg = super().format(record)
            self._style._fmt = original_format
            return mssg
        else:
            return super().format(record)

class FireXTaskFormatter(FireXFormatter):
    def format(self, record):
        task = get_current_task()
        if task and task.request:
            record.__dict__.update(task_id=task.request.id,
                                   task_name=task.name)
        else:
            record.__dict__.setdefault('task_name', '???')
            record.__dict__.setdefault('task_id', '???')
        return super().format(record)


@after_setup_task_logger.connect
def configure_task_logger(logger, loglevel, logfile, format, colorize, **_kwargs):
    # Find the WatchedFileHandler
    file_handler = [handler for handler in logger.handlers if isinstance(handler, WatchedFileHandler)][0]
    # set it's formatter to our custom Formatter
    file_handler.addFilter(AddHtmlElementsToLogRecords())
    file_handler.setFormatter(FireXTaskFormatter(format))

@after_setup_logger.connect
def configure_main_logger(logger, loglevel, logfile, format, colorize, **_kwargs):
    # Find the WatchedFileHandler
    file_handler = [handler for handler in logger.handlers if isinstance(handler, WatchedFileHandler)][0]
    # set it's formatter to our custom Formatter
    file_handler.addFilter(AddHtmlElementsToLogRecords())
    file_handler.setFormatter(FireXFormatter(format))
    # Deduce the worker name from the logfile, which is unfortunate
    worker_name = os.path.splitext(os.path.basename(logfile))[0]
    base_dir = os.path.dirname(logfile)
    logs_url = app.conf.logs_url
    if not logs_url:
        logs_url = os.path.relpath(app.conf.logs_dir, base_dir)

    html_header = JINJA_ENV.get_template('log_template.html').render(
        worker_log=True,
        firex_stylesheet=get_firex_css_filepath(app.conf.resources_dir, relative_from=base_dir),
        logo=get_firex_logo_filepath(app.conf.resources_dir, relative_from=base_dir),
        link_for_logo=app.conf.link_for_logo,
        header_main_title=worker_name,
        firex_id=app.conf.uid,
        logs_dir_url=logs_url)
    logger.raw(html_header)



import os
import sys
import inspect


def find_in_stack(file_to_find)->bool:
    frame = inspect.currentframe()
    while frame.f_back:
        if str(frame.f_code.co_filename).endswith(file_to_find):
            return True
        frame = frame.f_back
    return False


def is_running_under_coverage()->bool:
    return find_in_stack("coverage/cmdline.py")


def is_celery()->bool:
    return find_in_stack("celery/bin/celery.py")


def restart_celery_under_coverage():
    print("Restarting celery in coverage...", file=sys.stdout)

    # assemble new command line
    import coverage
    coverage_cmd = [sys.executable, os.path.dirname(coverage.__file__), "run", "--branch", "--parallel-mode", "-m"]
    coverage_celery = coverage_cmd + ["celery"] + sys.argv[1:]
    # coverage does not support forks
    coverage_celery +=["--pool=eventlet"]
    print(coverage_celery)

    # restart the process under coverage
    os.execlp(sys.executable, *coverage_celery)


if is_running_under_coverage():
    print("Currently running under coverage")
elif is_celery():
    restart_celery_under_coverage()



import os

from firexapp.broker_manager.redis_manager import RedisManager, RedisPasswordReadError
from firexapp.engine.celery import app

REDIS_BIN_ENV = "redis_bin_dir"


def get_redis_bin_dir():
    redis_bin_dir = os.environ.get(REDIS_BIN_ENV, "")
    if not redis_bin_dir:
        try:
            redis_bin_dir = app.conf.redis_bin_dir
        except AttributeError:
            pass
    return redis_bin_dir


class BrokerFactory:
    broker_env_variable = 'BROKER'

    @classmethod
    def set_broker_env(cls, broker_url):
        os.environ[cls.broker_env_variable] = broker_url

    @classmethod
    def create_new_broker_manager(cls, *args, **kwargs) -> RedisManager:
        return RedisManager(*args, redis_bin_base=get_redis_bin_dir(), **kwargs)

    @classmethod
    def broker_manager_from_env(cls) -> RedisManager:
        existing_broker_url = cls.get_broker_url(assert_if_not_set=True)
        hostname, port = RedisManager.get_hostname_port_from_url(existing_broker_url)
        password = RedisManager.get_password_from_url(existing_broker_url)
        return RedisManager(redis_bin_base=get_redis_bin_dir(),
                            hostname=hostname,
                            port=port,
                            password=password)

    @classmethod
    def broker_manager_from_logs_dir(cls, logs_dir, passwordless_fallback=False) -> RedisManager:
        hostname, port = RedisManager.get_hostname_port_from_logs_dir(logs_dir)
        try:
            password = RedisManager.get_password_from_logs_dir(logs_dir)
        except RedisPasswordReadError as e:
            if not passwordless_fallback:
                raise

            RedisManager.log(f'Cannot read previous broker password. Trying a new (random) password.', exc_info=e)
            # Setting this to None will cause the broker manager to create a new password
            password = None

        return RedisManager(redis_bin_base=get_redis_bin_dir(),
                            hostname=hostname,
                            port=port,
                            # FIXME: The RedisManager class offers different capabilities depending on how it's
                            #  initialized. Some functions don't work if it doesn't have a logs_dir.
                            logs_dir=logs_dir,
                            password=password)

    @classmethod
    def get_broker_url_from_logs_dir(cls, logs_dir) -> str:
        return RedisManager.get_broker_url_from_logs_dir(logs_dir)

    @classmethod
    def get_broker_url(cls, assert_if_not_set=False) -> str:
        url = os.environ.get(cls.broker_env_variable, "")
        if assert_if_not_set and not url:
            raise BrokerManagerException('%s env variable has not been set' % cls.broker_env_variable)
        return url

    @classmethod
    def get_broker_failed_auth_str(cls) -> str:
        return RedisManager.get_broker_failed_auth_str()

    @staticmethod
    def get_hostname_port_from_url(broker_url):
        return RedisManager.get_hostname_port_from_url(broker_url)


class BrokerManagerException(Exception):
    pass



from typing import NamedTuple, Optional
import json
from urllib.parse import urljoin, urlparse
import shutil
import os

from firexkit.resources import get_packaged_install_config_path
from firexapp.submit.uid import Uid
from firexapp.common import render_template

INSTALL_CONFIGS_ENV_NAME = 'firex_install_config'
INSTALL_CONFIGS_RUN_BASENAME = 'install-configs.json'

class FireXViewerTemplates(NamedTuple):
    viewer_base: str = ""
    run_path_template: str = ""
    task_path_template: str = ""
    run_logs_root_path_template: str = ""
    run_logs_entry_path_template: str = ""


# Data-only representation of the config. This is expected to EXACTLY reflect the contents of the config file.
# Utilities on top of this data should go in the FireXInstallConfigs class.
class FireXRawInstallConfigs(NamedTuple):
    viewer_templates: Optional[FireXViewerTemplates] = None

    # None means "use all installed tracking services". A list means only listed tracking services will be started,
    # and a service with each requested name must be installed. If a service is included here but not present during
    # start, the run will fail.
    requested_tracking_services: Optional[list] = None

    submit_args: Optional[dict] = None


class FireXInstallConfigError(Exception):
    pass


def install_config_path_from_logs_dir(logs_dir):
    return os.path.join(logs_dir, Uid.debug_dirname, INSTALL_CONFIGS_RUN_BASENAME)


def load_existing_raw_install_config(logs_dir) -> FireXRawInstallConfigs:
    install_config_path = install_config_path_from_logs_dir(logs_dir)
    try:
        with open(install_config_path) as fp:
            install_configs_dict = json.load(fp)
    except (OSError, json.JSONDecodeError) as e:
        raise FireXInstallConfigError(f"Failed to load install config from {install_config_path}") from e
    else:
        if install_configs_dict.get('viewer_templates'):
            viewer_config = FireXViewerTemplates(**install_configs_dict['viewer_templates'])
        else:
            viewer_config = None
        return FireXRawInstallConfigs(**{**install_configs_dict, 'viewer_templates': viewer_config})


class FireXInstallConfigs:
    """Utility functionality on top of data-only representation of configs."""

    def __init__(self, firex_id: str, logs_dir: str, raw_configs: FireXRawInstallConfigs):
        self.firex_id = firex_id
        self.logs_dir = logs_dir
        self.raw_configs = raw_configs
        self.run_url = self.get_run_url() if self.has_viewer() else None

    def has_viewer(self):
        return self.raw_configs.viewer_templates is not None

    def get_run_url(self) -> str:
        assert self.has_viewer(), "Callers must verify install configs specify URLs."
        return self._template_viewer_url(self.raw_configs.viewer_templates.run_path_template,
                                         {'firex_id': self.firex_id})

    def get_log_entry_url(self, log_entry_rel_run_root) -> str:
        assert self.has_viewer(), "Callers must verify install configs specify URLs."
        return self._template_viewer_url(self.raw_configs.viewer_templates.run_logs_entry_path_template,
                                         {'firex_id': self.firex_id,
                                          'run_logs_dir': self.logs_dir,
                                          'log_entry_rel_run_root': log_entry_rel_run_root})

    def get_logs_root_url(self) -> str:
        assert self.has_viewer(), "Callers must verify install configs specify URLs."
        return self._template_viewer_url(self.raw_configs.viewer_templates.run_logs_root_path_template,
                                         {'firex_id': self.firex_id,
                                          'run_logs_dir': self.logs_dir})

    def _template_viewer_url(self, template_str: str, template_args: dict) -> str:
        assert self.has_viewer(), "Callers must verify install configs specify URLs."

        rendered_template = render_template(template_str, template_args)
        parsed_template = urlparse(rendered_template)
        if parsed_template.scheme and parsed_template.netloc:
            # If the template can be parsed as a URL with scheme and netloc (host), it's already absolute,
            # so don't prepend base_url:
            return rendered_template
        # Assume rendered template is only path portion of URL and needs base prepended to become absolute.
        return urljoin(self.raw_configs.viewer_templates.viewer_base, rendered_template)

    def get_submit_args(self) -> dict:
        return self.raw_configs.submit_args


def load_existing_install_configs(firex_id: str, logs_dir: str) -> FireXInstallConfigs:
    return FireXInstallConfigs(firex_id, logs_dir, load_existing_raw_install_config(logs_dir))


# See https://stackoverflow.com/questions/33181170/how-to-convert-a-nested-namedtuple-to-a-dict/39235373
def isnamedtupleinstance(x):
    _type = type(x)
    bases = _type.__bases__
    if len(bases) != 1 or bases[0] != tuple:
        return False
    fields = getattr(_type, '_fields', None)
    if not isinstance(fields, tuple):
        return False
    return all(type(i)==str for i in fields)


def recursive_named_tuple_asdict(obj):
    if isinstance(obj, dict):
        return {key: recursive_named_tuple_asdict(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [recursive_named_tuple_asdict(value) for value in obj]
    elif isnamedtupleinstance(obj):
        return {key: recursive_named_tuple_asdict(value) for key, value in obj._asdict().items()}
    elif isinstance(obj, tuple):
        return tuple(recursive_named_tuple_asdict(value) for value in obj)
    else:
        return obj


def load_new_install_configs(firex_id: str, logs_dir: str, install_config_path: Optional[str],
                             raw_install_config: Optional[FireXRawInstallConfigs] = None) -> FireXInstallConfigs:
    """
    Copies supplied install configs to supplied logs_dir and returns loaded (i.e. deserialized)
    FireXInstallConfigs object. If no install_config_path is supplied, internally-defined default install config is
    used. Install config file is guaranteed to be written to logs_dir.
    """
    assert not (install_config_path and raw_install_config), \
        "Cannot specify both a file to load config from and an explicit config object."
    install_config_copy_path = install_config_path_from_logs_dir(logs_dir)

    # Either write configs or copy input file.
    if install_config_path is None:
        if raw_install_config is not None:
            raw_configs_to_write = raw_install_config
        else:
            # built-in default configs
            raw_configs_to_write = FireXRawInstallConfigs(viewer_templates=None, requested_tracking_services=None)
        with open(install_config_copy_path, 'w') as fp:
            json.dump(recursive_named_tuple_asdict(raw_configs_to_write), fp)
    else:
        # Copy supplied JSON file specifying config.
        try:
            if not os.path.isabs(install_config_path) and not os.path.isfile(install_config_path):
                # A non-absolute file that doesn't exist locally can be loaded from firexkit resources.
                resource_install_config = get_packaged_install_config_path(install_config_path)
                if os.path.isfile(resource_install_config):
                    install_config_path = resource_install_config
            shutil.copyfile(install_config_path, install_config_copy_path)
        except OSError as e:
            raise FireXInstallConfigError(f"Failed to load install config from {install_config_path}") from e

    return load_existing_install_configs(firex_id, logs_dir)



import os
import argparse
import logging
import subprocess
import time
from psutil import Process, TimeoutExpired
from collections import namedtuple

from celery import Celery

from firexapp.celery_manager import CeleryManager
from firexapp.submit.uid import Uid
from firexapp.broker_manager.broker_factory import BrokerFactory, REDIS_BIN_ENV
from firexkit.inspect import get_active, get_revoked
from firexapp.common import qualify_firex_bin, select_env_vars

logger = logging.getLogger(__name__)


CELERY_SHUTDOWN_WAIT = 5 * 60
MaybeCeleryActiveTasks = namedtuple('MaybeCeleryActiveTasks', ['celery_read_success', 'active_tasks'])


def launch_background_shutdown(logs_dir, reason):
    try:
        shutdown_cmd = [qualify_firex_bin("firex_shutdown"), "--logs_dir",  logs_dir]
        if reason:
            shutdown_cmd += ['--reason', reason]
        pid = subprocess.Popen(shutdown_cmd, close_fds=True, env=select_env_vars([REDIS_BIN_ENV, 'PATH'])).pid
    except Exception as e:
        logger.error("SHUTDOWN PROCESS FAILED TO LAUNCH -- REDIS WILL LEAK.")
        logger.error(e)
        raise
    else:
        try:
            Process(pid).wait(0.1)
        except TimeoutExpired:
            logger.debug("Started background shutdown with pid %s" % pid)
        else:
            logger.error("SHUTDOWN PROCESS FAILED TO RUN -- REDIS WILL LEAK.")


def wait_for_broker_shutdown(broker, timeout=15, force_kill=True):
    logger.debug("Waiting for broker to shut down")
    shutdown_wait_time = time.time() + timeout
    while time.time() < shutdown_wait_time:
        if not broker.is_alive():
            break
        time.sleep(0.1)

    if not broker.is_alive():
        logger.debug("Confirmed successful graceful broker shutdown.")
    elif force_kill:
        logger.debug("Warning! Broker was not shut down after %s seconds. FORCE KILLING BROKER." % str(timeout))
        broker.force_kill()

    return not broker.is_alive()


def get_active_broker_safe(broker, celery_app):
    #
    # Note: get_active can hang in celery/kombu library if broker is down.
    # Checking for broker.is_alive() is an attempt to prevent that, but note
    # the broker can theoretically die between that call and the get_active() call.
    if not broker.is_alive():
        return None
    return get_active(inspect_retry_timeout=4, celery_app=celery_app)


def get_revoked_broker_safe(broker, celery_app):
    if not broker.is_alive():
        return None
    return get_revoked(inspect_retry_timeout=4, celery_app=celery_app)


def _tasks_from_active(active, task_predicate) -> MaybeCeleryActiveTasks:
    if active is None:
        return MaybeCeleryActiveTasks(False, None)
    tasks = []
    for host in active.values():
        tasks.extend([t for t in host if task_predicate(t)])
    return MaybeCeleryActiveTasks(True, tasks)


def revoke_active_tasks(broker, celery_app,  max_revoke_retries=5, task_predicate=lambda task: True):
    logger.debug("Querying Celery to find any remaining active tasks.")
    maybe_active_tasks = _tasks_from_active(get_active_broker_safe(broker, celery_app), task_predicate)
    revoke_retries = 0
    # Revoke retry loop
    while (maybe_active_tasks.celery_read_success
           and maybe_active_tasks.active_tasks
           and revoke_retries < max_revoke_retries):
        if revoke_retries:
            logger.warning("Found %s active tasks after revoke. Revoking active tasks again."
                           % len(maybe_active_tasks.active_tasks))

        # Revoke tasks in order they were started. This avoids ChainRevokedException errors when children are revoked
        # before their parents.
        for task in sorted(maybe_active_tasks.active_tasks, key=lambda t: t.get('time_start', float('inf'))):
            logger.info(f"Revoking {task['name']}[{task['id']}]")
            celery_app.control.revoke(task_id=task["id"], terminate=True)

        # wait for confirmation of revoke
        maybe_active_tasks = _tasks_from_active(get_active_broker_safe(broker, celery_app), task_predicate)
        wait_for_task_revoke_start = time.monotonic()
        while maybe_active_tasks.celery_read_success and maybe_active_tasks.active_tasks \
                and time.monotonic() - wait_for_task_revoke_start < 3:
            time.sleep(0.25)
            maybe_active_tasks = _tasks_from_active(get_active_broker_safe(broker, celery_app), task_predicate)

        revoke_retries += 1

    if not maybe_active_tasks.celery_read_success:
        logger.info("Failed to read active tasks from celery. May shutdown with unrevoked tasks.")
    elif len(maybe_active_tasks.active_tasks) == 0:
        logger.info("Confirmed no active tasks after revoke.")
    elif revoke_retries >= max_revoke_retries:
        logger.warning("Exceeded max revoke retry attempts, %s active tasks may not be revoked."
                       % len(maybe_active_tasks.active_tasks))


def init():
    parser = argparse.ArgumentParser()
    parser.add_argument("--logs_dir", help="Logs directory for the firexapp run to shutdown.",
                        required=True)
    parser.add_argument("--reason", help="A reason that will be logged for clarity.",
                        required=False, default='No reason provided.')
    args = parser.parse_args()
    logs_dir = args.logs_dir

    log_file = os.path.join(logs_dir, Uid.debug_dirname, 'shutdown.log.txt')
    logging.basicConfig(filename=log_file, level=logging.DEBUG,
                        format='[%(asctime)s %(levelname)s] %(message)s',
                        datefmt="%Y-%m-%d %H:%M:%S")

    return logs_dir, args.reason


def shutdown_run(logs_dir, reason='No reason provided'):
    logger.info(f"Shutting down due to reason: {reason}")
    logger.info(f"Shutting down with logs: {logs_dir}.")
    broker = BrokerFactory.broker_manager_from_logs_dir(logs_dir)
    logger.info(f"Shutting down with broker: {broker.broker_url_safe_print}.")

    celery_manager = CeleryManager(logs_dir=logs_dir, broker=broker)
    celery_app = Celery(broker=broker.broker_url,
                        accept_content=['pickle', 'json'])

    try:
        if get_active_broker_safe(broker, celery_app):
            revoke_active_tasks(broker, celery_app)

            logger.info("Found active Celery; sending Celery shutdown.")
            celery_app.control.shutdown()

            celery_shutdown_success = celery_manager.wait_for_shutdown(CELERY_SHUTDOWN_WAIT)
            if not celery_shutdown_success:
                logger.warning(f"Celery not shutdown after {CELERY_SHUTDOWN_WAIT} secs, force killing instead.")
                celery_manager.shutdown()
            else:
                logger.debug("Confirmed Celery shutdown successfully.")
        elif celery_manager.find_all_procs():
            logger.info("Celery not active, but found celery processes to force shutdown.")
            celery_manager.shutdown()
        else:
            logger.info("No active Celery processes.")
    finally:
        if broker.is_alive():
            logger.info("Broker is alive; sending redis shutdown.")
            # broker will be shut down by celery if active
            broker.shutdown()
            wait_for_broker_shutdown(broker)
        else:
            logger.info("No active Broker.")

    logger.info("Shutdown of %s complete." % logs_dir)


def main():
    logs_dir, reason = init()
    try:
        shutdown_run(logs_dir, reason)
    except Exception as e:
        logger.exception(e)



from celery.utils.log import get_task_logger
from firexkit.chain import FireXTask, InjectArgs
from firexkit.task import flame, flame_collapse

from firexapp.engine.celery import app

logger = get_task_logger(__name__)

# noinspection PyPep8Naming
@app.task(bind=True, returns=FireXTask.DYNAMIC_RETURN)
@flame('status')
def CopyBogKeys(self: FireXTask, bog_key_map: dict, strict: bool = False):
    """
    This service copies selected keys from this FireXTask instance's bog into new keys with a different name and
    returns the resulting dictionary. It can therefore be used to preserve values in a chain when they will
    be trampled by return values from downstream in the chain.

    :param bog_key_map: mapping from existing key names to new key names that existing bog values should be copied to
        in the return value.
    :param strict: [True|False] (default=False):
        True: All source entries specified in the mapping must exist in the  bog or the service fails.
        False: Skip source entries that don't exist in bog.
    :return: dict with keys from the values of bog_key_map, and values from the BoG.
    """

    logger.debug('abog content: %r' % self.abog)
    flame_status = ""

    new = {}
    for existing_bog_key, new_key in bog_key_map.items():
        try:
            existing_value = self.abog[existing_bog_key]
        except KeyError:
            if strict:
                raise AssertionError(f'Strict is specified and no entry found for "{existing_bog_key}" in bog.')
            logger.debug(f'No entry for "{existing_bog_key}" in bog. Skipping...')
        else:
            new[new_key] = existing_value
            status_str = f'{existing_bog_key}={new_key}'
            logger.debug('BOG mapping: ' + status_str)
            if flame_status:
                flame_status += '<BR>'
            flame_status += status_str
            self.send_firex_html(status=flame_status)

    return new


# noinspection PyPep8Naming
@app.task(bind=True, returns=FireXTask.DYNAMIC_RETURN)
@flame_collapse('self')
def ScheduleSubChain(self: FireXTask, chain, enqueue_args: dict = {}, catch_errors: bool = False,
                     inject_abog: bool = True):
    """
    This service will schedule the chain provided and return all values returned by this chain.

    :chain: Chain of services to schedule.
    :enqueue_args: Args to propagate to the enqueue call
    :catch_errors: True=Catch errors generated by the chain and  return success anyway.
                    (return values will {} in this case).
                   False=Exceptions in the chain will trigger an exception in this service as well.

    :inject_abog: True=(Default) Inject the abog into the subchain
                  False=Run subchain without abog injection

    :return: All values returned by the scheduled chain.
    """

    results = {}

    sub_chain = chain
    if inject_abog:
        sub_chain = InjectArgs(**self.abog) | chain

    try:
        results = self.enqueue_child_and_get_results(sub_chain, **enqueue_args)
    except Exception as e:
        if catch_errors:
            logger.error(f'Sub-chain raised an exception:\n{e}', exc_info=e)
        else:
            raise

    return results



from firexapp.submit.console import setup_console_logging
from firexapp.submit.uid import Uid
from logging import INFO, DEBUG, WARNING
import os
import re
import subprocess
import psutil
from firexapp.broker_manager.broker_factory import BrokerFactory
from socket import gethostname
from firexapp.common import poll_until_file_not_empty, poll_until_dir_empty, find_procs
from firexapp.plugins import PLUGGING_ENV_NAME, cdl2list
from firexapp.fileregistry import FileRegistry
from collections.abc import Iterable

logger = setup_console_logging(__name__)

CELERY_LOGS_REGISTRY_KEY = 'celery_logs'
FileRegistry().register_file(CELERY_LOGS_REGISTRY_KEY, os.path.join(Uid.debug_dirname, 'celery'))

CELERY_PIDS_REGISTRY_KEY = 'celery_pids'
FileRegistry().register_file(CELERY_PIDS_REGISTRY_KEY,
                             os.path.join(FileRegistry().get_relative_path(CELERY_LOGS_REGISTRY_KEY), 'pids'))

MICROSERVICE_LOGS_REGISTRY_KEY = 'microservice_logs'
FileRegistry().register_file(MICROSERVICE_LOGS_REGISTRY_KEY, 'microservice_logs')


class CeleryWorkerStartFailed(Exception):
    pass

class CeleryManager(object):
    celery_bin_name = 'celery'

    def __init__(self, plugins=None, logs_dir=None, worker_log_level='debug', cap_concurrency=None,
                 app='firexapp.engine', celery_bin_dir='', env=None, broker=None):

        if not broker:
            self.broker = BrokerFactory.get_broker_url(assert_if_not_set=True)
        self.hostname = gethostname()
        self.plugins = plugins
        self.logs_dir = logs_dir
        self.worker_log_level = worker_log_level
        self.cap_concurrency = cap_concurrency
        self.app = app
        self.celery_bin_dir = celery_bin_dir

        self.env = os.environ.copy()
        self.update_env(self.get_plugins_env(plugins))
        if env:
            self.update_env(env)

        self.pid_files = dict()
        self._celery_logs_dir = None
        self._celery_pids_dir = None
        self._workers_logs_dir = None

    @property
    def celery_bin(self):
        return os.path.join(self.celery_bin_dir, self.celery_bin_name)

    @classmethod
    def log(cls, msg, header=None, level=DEBUG):
        if header is None:
            header = cls.__name__
        if header:
            msg = '[%s] %s' % (header, msg)
        logger.log(level, msg)

    def update_env(self, env):
        assert isinstance(env, dict), 'env needs to be a dictionary'
        self.env.update({k: str(v) for k, v in env.items()})

    @staticmethod
    def get_plugins_env(plugins):
        plugin_files = cdl2list(plugins)
        return {PLUGGING_ENV_NAME: ",".join(plugin_files)}

    @staticmethod
    def get_celery_logs_dir(logs_dir):
        return FileRegistry().get_file(CELERY_LOGS_REGISTRY_KEY, logs_dir)

    @staticmethod
    def get_celery_pids_dir(logs_dir):
        return FileRegistry().get_file(CELERY_PIDS_REGISTRY_KEY, logs_dir)

    @staticmethod
    def get_worker_logs_dir(logs_dir):
        return FileRegistry().get_file(MICROSERVICE_LOGS_REGISTRY_KEY, logs_dir)

    @property
    def celery_logs_dir(self):
        if not self._celery_logs_dir:
            _celery_logs_dir = self.get_celery_logs_dir(self.logs_dir)
            os.makedirs(_celery_logs_dir, exist_ok=True)
            self._celery_logs_dir = _celery_logs_dir
        return self._celery_logs_dir

    @property
    def celery_pids_dir(self):
        if not self._celery_pids_dir:
            _celery_pids_dir = self.get_celery_pids_dir(self.logs_dir)
            os.makedirs(_celery_pids_dir, exist_ok=True)
            self._celery_pids_dir = _celery_pids_dir
        return self._celery_pids_dir

    @property
    def workers_logs_dir(self):
        if not self._workers_logs_dir:
            _workers_logs_dir = self.get_worker_logs_dir(self.logs_dir)
            os.makedirs(_workers_logs_dir, exist_ok=True)
            self._workers_logs_dir = _workers_logs_dir
        return self._workers_logs_dir

    @classmethod
    def get_worker_log_file(cls, logs_dir, worker_and_host):
        return cls.__get_worker_log_file(cls.get_worker_logs_dir(logs_dir), worker_and_host)

    def _get_worker_log_file(self, workername):
        queue_and_worker = self.get_worker_and_host(workername, self.hostname)
        return self.__get_worker_log_file(self.workers_logs_dir, queue_and_worker)

    @staticmethod
    def __get_worker_log_file(worker_logs_dir, worker_and_host):
        return os.path.join(worker_logs_dir, '%s.html' % worker_and_host)

    @classmethod
    def get_pid_file(cls, logs_dir, workername, hostname=gethostname()):
        worker_and_host = cls.get_worker_and_host(workername, hostname)
        return cls.__get_pid_file(cls.get_celery_pids_dir(logs_dir), worker_and_host)

    def _get_pid_file(self, workername):
        worker_and_host = self.get_worker_and_host(workername, self.hostname)
        return self.__get_pid_file(self.celery_pids_dir, worker_and_host)

    @staticmethod
    def __get_pid_file(pids_logs_dir, worker_and_host):
        return os.path.join(pids_logs_dir, '%s.pid' % worker_and_host)

    def _get_stdout_file(self, workername):
        return os.path.join(self.celery_logs_dir, '%s@%s.stdout.txt' % (workername, self.hostname))

    @staticmethod
    def get_worker_and_host(workername, hostname):
        return '%s@%s' % (workername, hostname)

    @classmethod
    def get_pid_from_file(cls, pid_file):
        try:
            with open(pid_file) as f:
                pid = f.read().strip()
        except FileNotFoundError:
            cls.log('No pid file found in %s' % pid_file, level=WARNING)
            raise
        else:
            if pid:
                return int(pid)
            else:
                raise AssertionError('no pid')

    @classmethod
    def get_pid(cls, logs_dir, workername, hostname=gethostname()):
        pid_file = cls.get_pid_file(logs_dir, workername, hostname)
        return cls.get_pid_from_file(pid_file)

    @classmethod
    def get_worker_pids(cls, logs_dir, hostname, workernames):
        hostname = gethostname() if hostname == 'localhost' else hostname
        pids = []
        for workername in workernames:
            try:
                pid = cls.get_pid(logs_dir, workername, hostname)
            except Exception as e:
                cls.log(e)
            else:
                pids.append(pid)
        return pids

    @staticmethod
    def cap_cpu_count(count, cap_concurrency):
        return min(count, cap_concurrency) if cap_concurrency else count

    def extract_errors_from_celery_logs(self, celery_log_file, max_errors=20):
        err_list = None
        try:
            with open(celery_log_file, encoding='ascii', errors='ignore') as f:
                logs = f.read()
                err_list = re.findall(r'^\S*Error: .*$', logs, re.MULTILINE)
                if err_list:
                    err_list = err_list[0:max_errors]
        except FileNotFoundError:
            pass

        return err_list

    def wait_until_active(self, pid_file, stdout_file, workername, timeout=15*60):
        extra_err_info = ''
        try:
            poll_until_file_not_empty(pid_file, timeout=timeout)
        except AssertionError:
            err_list = self.extract_errors_from_celery_logs(stdout_file)
            if err_list:
                extra_err_info += '\nFound the following errors:\n' + '\n'.join(err_list)

            raise CeleryWorkerStartFailed('The worker %s@%s did not come up after %d seconds.'
                                          '\nPlease look into %s for details.'
                                          '%s' % (workername,
                                                  self.hostname,
                                                  timeout,
                                                  stdout_file,
                                                  extra_err_info))
        pid = self.get_pid_from_file(pid_file)
        self.log('pid %d became active' % pid)

    def start(self, workername, queues=None, wait=True, timeout=15*60, concurrency=None, worker_log_level=None,
              app=None, cap_concurrency=None, cwd=None, soft_time_limit=None, autoscale: tuple = None):

        # Override defaults if applicable
        worker_log_level = worker_log_level if worker_log_level else self.worker_log_level
        app = app if app else self.app
        cap_concurrency = cap_concurrency if cap_concurrency else self.cap_concurrency

        stdout_file = self._get_stdout_file(workername)
        log_file = self._get_worker_log_file(workername)
        pid_file = self._get_pid_file(workername)
        self.pid_files[workername] = pid_file

        cmd = '%s --app=%s worker --hostname=%s@%%h --loglevel=%s ' \
              '--logfile=%s --pidfile=%s --events --without-gossip --without-heartbeat --without-mingle -Ofair' % \
              (self.celery_bin, app, workername, worker_log_level, log_file, pid_file)
        if queues:
            cmd += ' --queues=%s' % queues

        if concurrency and autoscale:
            raise AssertionError('You can either provide a value of concurrency or autoscale, but not both')

        if concurrency:
            cmd += ' --concurrency=%d' % self.cap_cpu_count(concurrency, cap_concurrency)
        elif autoscale:
            assert isinstance(autoscale, Iterable), 'autoscale should be a tuple of (min, max)'
            assert len(autoscale) == 2, 'autoscale should be a tuple of two elements (min, max)'
            autoscale_v1, autoscale_v2 = autoscale
            autoscale_min = min(autoscale_v1, autoscale_v2)
            autoscale_max = max(autoscale_v1, autoscale_v2)
            autoscale_min = self.cap_cpu_count(autoscale_min, cap_concurrency)
            autoscale_max = self.cap_cpu_count(autoscale_max, cap_concurrency)
            cmd += f' --autoscale={autoscale_max},{autoscale_min}'
        if soft_time_limit:
            cmd += f' --soft-time-limit={soft_time_limit}'

        # piping to ts is helpful for debugging if available
        try:
            subprocess.check_call(["which", "ts"], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
        except subprocess.CalledProcessError:
            pass
        else:
            cmd += " | ts '[%Y-%m-%d %H:%M:%S]'"
        cmd += ' &'

        self.log('Starting %s on %s...' % (workername, self.hostname))
        self.log(cmd)

        if cwd:
            self.log('cwd=%s' % cwd)

        with open(stdout_file, 'ab') as fp:
            subprocess.check_call(cmd, shell=True, stdout=fp, stderr=subprocess.STDOUT, env=self.env,
                                  cwd=cwd)

        if wait:
            self.wait_until_active(pid_file=pid_file, timeout=timeout, stdout_file=stdout_file, workername=workername)

    @staticmethod
    def find_procs(pid_file):
        return find_procs('celery', cmdline_contains='--pidfile=%s' % pid_file)

    def find_all_procs(self):
        procs = []
        for pid_file in os.listdir(self.celery_pids_dir):
            procs += self.find_procs(os.path.join(self.celery_pids_dir, pid_file))
        return procs

    def kill_all_forked(self, pid_file):
        for proc in self.find_procs(pid_file):
            self.log('Killing  pid %d' % proc.pid, level=INFO)
            try:
                proc.kill()
            except Exception:
                self.log('Failed to kill pid %d' % proc.pid, level=WARNING)

    @classmethod
    def terminate(cls, pid, timeout=60):
        cls.log('Terminating pid %d' % pid, level=INFO)
        p = psutil.Process(pid)
        p.terminate()
        p.wait(timeout=timeout)

    def shutdown(self, timeout=60):
        if self.pid_files:
            name_to_pid_file = self.pid_files
        else:
            # self.pid_files is only populated when starting celery, so if this manager didn't start the celery
            # instance being operated on, fallback to the pid directory.
            name_to_pid_file = {pf: os.path.join(self.celery_pids_dir, pf) for pf in os.listdir(self.celery_pids_dir)}

        for name, pid_file in name_to_pid_file.items():
            self.log('Attempting shutdown of %s' % name)
            try:
                pid = self.get_pid_from_file(pid_file)
            except Exception as e:
                self.log(e)
            else:
                try:
                    self.terminate(pid, timeout=timeout)
                except (psutil.TimeoutExpired, psutil.NoSuchProcess):
                    self.kill_all_forked(pid_file)
                except Exception as e:
                    self.log(e)

    def wait_for_shutdown(self, timeout=15):
        return poll_until_dir_empty(self.celery_pids_dir, timeout=timeout)



# This module serves as a Singleton that will store
# a registry of the ouptut files needed
import json
from firexapp.submit.uid import Uid

import os


class KeyAlreadyRegistered(Exception):
    pass


class KeyNotRegistered(Exception):
    pass


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class FileRegistry(metaclass=Singleton):
    def __init__(self, from_file=None):
        if from_file:
            self.file_registry = self.read_from_file(from_file)
        else:
            self.file_registry = {}

    @classmethod
    def destroy(cls):
        cls._instances = {}

    def register_file(self, key, relative_path):
        if key in self.file_registry:
            raise KeyAlreadyRegistered('%r already registered; callable=%s' % (key, self.file_registry[key]))
        else:
            self.file_registry[key] = relative_path

    def get_file(self, key, uid_or_logsdir):
        try:
            return self.resolve_path(uid_or_logsdir, self.get_relative_path(key))
        except KeyError:
            raise KeyNotRegistered('%r is not registered' % key)

    def get_relative_path(self, key):
        return self.file_registry[key]

    @staticmethod
    def resolve_path(uid_or_logsdir, relative_path):
        logs_dir = uid_or_logsdir.logs_dir if isinstance(uid_or_logsdir, Uid) else uid_or_logsdir
        return os.path.join(logs_dir, relative_path)

    @staticmethod
    def read_from_file(path):
        with open(path) as fp:
            return json.load(fp)

    def dump_to_file(self, path):
        with open(path, 'w') as fp:
            json.dump(self.file_registry, fp, sort_keys=True, indent=2)



import os

from celery.app.base import Celery
from firexkit.task import FireXTask
from celery.signals import celeryd_init
from firexapp.submit.install_configs import install_config_path_from_logs_dir, load_existing_install_configs


firex_app_name = '.'.join(__name__.split(".")[:-1])
app = Celery(strict_typing=False, task_cls='%s:%s' % (FireXTask.__module__, FireXTask.__name__))
app.config_from_object(firex_app_name + '.default_celery_config')


@celeryd_init.connect
def add_items_to_conf(conf=None, **_kwargs):
    conf.uid = app.backend.get('uid').decode()
    conf.logs_dir = app.backend.get('logs_dir').decode()
    conf.link_for_logo = conf.logs_dir
    conf.logs_url = None
    conf.resources_dir = app.backend.get('resources_dir').decode()

    install_config = install_config_path_from_logs_dir(conf.logs_dir)
    assert os.path.isfile(install_config), \
        f"Install config missing from run, firexapp submit is expected to have populated it: {install_config}"

    # TODO: assumes everywhere celery is started can load from logs_dir. Should likely serialize to backend.
    conf.install_config = load_existing_install_configs(conf.uid, conf.logs_dir)
    if conf.install_config.has_viewer():
        conf.logs_url = conf.link_for_logo = conf.install_config.get_logs_root_url()
        conf.link_for_logo = conf.install_config.get_logs_root_url()



from collections import namedtuple
from enum import Enum
import logging
import re

from firexkit.result import ChainInterruptedException

logger = logging.getLogger(__name__)


# event entry key name that allows a task to indicate a child-like relationship with another task.
# For example, a delayed-dependency that initiates a non-child task to run can indicate it is an ancestor
# of that triggered non-child task, since it effectively cause it to execute, much like an ordinary (i.e. celery)
# parent task.
ADDITIONAL_CHILDREN_KEY = 'additional_children'
EXTERNAL_COMMANDS_KEY = 'external_commands'



class RunStates(Enum):
    RECEIVED = "task-received"
    STARTED = "task-started"
    BLOCKED = "task-blocked"
    UNBLOCKED = "task-unblocked"
    SUCCEEDED = "task-succeeded"
    FAILED = "task-failed"
    REVOKED = "task-revoked"
    INCOMPLETE = "task-incomplete"


ALL_RUNSTATES = {
    RunStates.RECEIVED.value: {'terminal': False},
    RunStates.STARTED.value: {'terminal': False},
    RunStates.BLOCKED.value: {'terminal': False},
    RunStates.UNBLOCKED.value: {'terminal': False},
    RunStates.SUCCEEDED.value: {'terminal': True},
    RunStates.FAILED.value: {'terminal': True},
    RunStates.REVOKED.value: {'terminal': True},
    RunStates.INCOMPLETE.value: {'terminal': True},  # server-side kludge state to fix tasks that will never complete.
}
COMPLETE_RUNSTATES = [s for s, v in ALL_RUNSTATES.items() if v['terminal']]
INCOMPLETE_RUNSTATES = [s for s, v in ALL_RUNSTATES.items() if not v['terminal']]


class RunMetadataColumn(Enum):
    FIREX_ID = "firex_id"
    LOGS_DIR = "logs_dir"
    CHAIN = "chain"
    ROOT_UUID = "root_uuid"


FireXRunMetadata = namedtuple('RunMetadata', [rmc.value for rmc in RunMetadataColumn])


# Note field order matters. TaskColumn is the authority on field order.
class TaskColumn(Enum):
    UUID = "uuid"
    FIREX_ID = "firex_id"
    CHAIN_DEPTH = "chain_depth"
    BOUND_ARGS = "firex_bound_args"
    RESULTS = "firex_result"
    DEFAULT_BOUND_ARGS = "firex_default_bound_args"
    FROM_PLUGIN = "from_plugin"
    HOSTNAME = "hostname"
    LOGS_URL = "logs_url"
    LONG_NAME = "long_name"
    NAME = "name"
    ACTUAL_RUNTIME = "actual_runtime"
    FIRST_STARTED = "first_started"
    PARENT_ID = "parent_id"
    RETRIES = "retries"
    STATE = "state"
    TASK_NUM = "task_num"
    UTCOFFSET = "utcoffset"
    EXCEPTION = "exception"
    TRACEBACK = "traceback"
    EXCEPTION_CAUSE_UUID = 'exception_cause_uuid'


TASK_COLUMN_NAMES = [tc.value for tc in TaskColumn]


def get_task_data(input_dict):
    return {k: v for k, v in input_dict.items() if k in TASK_COLUMN_NAMES}


FireXTask = namedtuple('FireXTask', [tc.value for tc in TaskColumn])


def is_chain_exception(task):
    return task.exception and task.exception.strip().startswith(ChainInterruptedException.__name__)


def get_chain_exception_child_uuid(task):
    assert is_chain_exception(task)
    exception_str = task.exception.strip()
    # example: ChainInterruptedException('ad9b0b79-86e9-4d76-8654-9c19886d50a1', ...).
    m = re.search(r'' + ChainInterruptedException.__name__ + "\('([\da-f\-]+)'", exception_str)
    assert m, "No UUID found in %s." % exception_str
    return m.group(1)


def is_failed(task: FireXTask, ignore_chain_exception=False):
    is_failure = task.state == RunStates.FAILED.value
    if not is_failure:
        return False

    if ignore_chain_exception:
        return not is_chain_exception(task)

    return True
